# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2017-2019 Unrud <unrud@outlook.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Radicale tests with simple requests and rights.
"""

import os

from radicale.tests import BaseTest
from radicale.tests.helpers import get_file_content


class TestBaseRightsRequests(BaseTest):
    """Tests basic requests with rights."""

    def _test_rights(self, rights_type: str, user: str, path: str, mode: str,
                     expected_status: int, with_auth: bool = True) -> None:
        assert mode in ("r", "w")
        assert user in ("", "tmp", "user@domain.test")
        htpasswd_file_path = os.path.join(self.colpath, ".htpasswd")
        with open(htpasswd_file_path, "w") as f:
            f.write("tmp:bepo\nother:bepo\nuser@domain.test:bepo")
        self.configure({
            "rights": {"type": rights_type},
            "auth": {"type": "htpasswd" if with_auth else "none",
                     "htpasswd_filename": htpasswd_file_path,
                     "htpasswd_encryption": "plain"}})
        for u in ("tmp", "other"):
            # Indirect creation of principal collection
            self.propfind("/%s/" % u, login="%s:bepo" % u)
        os.makedirs(os.path.join(self.colpath, "collection-root", "domain.test"), exist_ok=True)
        (self.propfind if mode == "r" else self.proppatch)(
            path, check=expected_status, login="%s:bepo" % user if user else None)

    def test_owner_only(self) -> None:
        self._test_rights("owner_only", "", "/", "r", 401)
        self._test_rights("owner_only", "", "/", "w", 401)
        self._test_rights("owner_only", "", "/tmp/", "r", 401)
        self._test_rights("owner_only", "", "/tmp/", "w", 401)
        self._test_rights("owner_only", "tmp", "/", "r", 207)
        self._test_rights("owner_only", "tmp", "/", "w", 403)
        self._test_rights("owner_only", "tmp", "/tmp/", "r", 207)
        self._test_rights("owner_only", "tmp", "/tmp/", "w", 207)
        self._test_rights("owner_only", "tmp", "/other/", "r", 403)
        self._test_rights("owner_only", "tmp", "/other/", "w", 403)

    def test_owner_only_without_auth(self) -> None:
        self._test_rights("owner_only", "", "/", "r", 207, False)
        self._test_rights("owner_only", "", "/", "w", 401, False)
        self._test_rights("owner_only", "", "/tmp/", "r", 207, False)
        self._test_rights("owner_only", "", "/tmp/", "w", 207, False)

    def test_owner_write(self) -> None:
        self._test_rights("owner_write", "", "/", "r", 401)
        self._test_rights("owner_write", "", "/", "w", 401)
        self._test_rights("owner_write", "", "/tmp/", "r", 401)
        self._test_rights("owner_write", "", "/tmp/", "w", 401)
        self._test_rights("owner_write", "tmp", "/", "r", 207)
        self._test_rights("owner_write", "tmp", "/", "w", 403)
        self._test_rights("owner_write", "tmp", "/tmp/", "r", 207)
        self._test_rights("owner_write", "tmp", "/tmp/", "w", 207)
        self._test_rights("owner_write", "tmp", "/other/", "r", 207)
        self._test_rights("owner_write", "tmp", "/other/", "w", 403)

    def test_owner_write_without_auth(self) -> None:
        self._test_rights("owner_write", "", "/", "r", 207, False)
        self._test_rights("owner_write", "", "/", "w", 401, False)
        self._test_rights("owner_write", "", "/tmp/", "r", 207, False)
        self._test_rights("owner_write", "", "/tmp/", "w", 207, False)

    def test_authenticated(self) -> None:
        self._test_rights("authenticated", "", "/", "r", 401)
        self._test_rights("authenticated", "", "/", "w", 401)
        self._test_rights("authenticated", "", "/tmp/", "r", 401)
        self._test_rights("authenticated", "", "/tmp/", "w", 401)
        self._test_rights("authenticated", "tmp", "/", "r", 207)
        self._test_rights("authenticated", "tmp", "/", "w", 207)
        self._test_rights("authenticated", "tmp", "/tmp/", "r", 207)
        self._test_rights("authenticated", "tmp", "/tmp/", "w", 207)
        self._test_rights("authenticated", "tmp", "/other/", "r", 207)
        self._test_rights("authenticated", "tmp", "/other/", "w", 207)

    def test_authenticated_without_auth(self) -> None:
        self._test_rights("authenticated", "", "/", "r", 207, False)
        self._test_rights("authenticated", "", "/", "w", 207, False)
        self._test_rights("authenticated", "", "/tmp/", "r", 207, False)
        self._test_rights("authenticated", "", "/tmp/", "w", 207, False)

    def test_from_file(self) -> None:
        rights_file_path = os.path.join(self.colpath, "rights")
        with open(rights_file_path, "w") as f:
            f.write("""\
[owner]
user: .+
collection: {user}(/.*)?
permissions: RrWw
[custom]
user: .*
collection: custom(/.*)?
permissions: Rr
[read-domain-principal]
user: .+@([^@]+)
collection: {0}
permissions: R""")
        self.configure({"rights": {"file": rights_file_path}})
        self._test_rights("from_file", "", "/other/", "r", 401)
        self._test_rights("from_file", "tmp", "/tmp/", "r", 207)
        self._test_rights("from_file", "tmp", "/other/", "r", 403)
        self._test_rights("from_file", "", "/custom/sub", "r", 404)
        self._test_rights("from_file", "tmp", "/custom/sub", "r", 404)
        self._test_rights("from_file", "", "/custom/sub", "w", 401)
        self._test_rights("from_file", "tmp", "/custom/sub", "w", 403)
        self._test_rights("from_file", "tmp", "/custom/sub", "w", 403)
        self._test_rights("from_file", "user@domain.test", "/domain.test/", "r", 207)
        self._test_rights("from_file", "user@domain.test", "/tmp/", "r", 403)
        self._test_rights("from_file", "user@domain.test", "/other/", "r", 403)

    def test_from_file_limited_get(self):
        rights_file_path = os.path.join(self.colpath, "rights")
        with open(rights_file_path, "w") as f:
            f.write("""\
[write-all]
user: tmp
collection: .*
permissions: RrWw
[limited-public]
user: .*
collection: public/[^/]*
permissions: i""")
        self.configure({"rights": {"type": "from_file",
                                   "file": rights_file_path}})
        self.configure({"auth": {"type": "none"}})
        self.mkcalendar("/tmp/calendar", login="tmp:bepo")
        self.mkcol("/public", login="tmp:bepo")
        self.mkcalendar("/public/calendar", login="tmp:bepo")
        self.get("/tmp/calendar", check=401)
        self.get("/public/", check=401)
        self.get("/public/calendar")
        self.get("/public/calendar/1.ics", check=401)

    def test_custom(self) -> None:
        """Custom rights management."""
        self._test_rights("radicale.tests.custom.rights", "", "/", "r", 401)
        self._test_rights(
            "radicale.tests.custom.rights", "", "/tmp/", "r", 207)

    def test_collections_and_items(self) -> None:
        """Test rights for creation of collections, calendars and items.

        Collections are allowed at "/" and "/.../".
        Calendars/Address books are allowed at "/.../.../".
        Items are allowed at "/.../.../...".

        """
        self.configure({"auth": {"type": "none"}})
        self.mkcalendar("/", check=401)
        self.mkcalendar("/user/", check=401)
        self.mkcol("/user/")
        self.mkcol("/user/calendar/", check=401)
        self.mkcalendar("/user/calendar/")
        self.mkcol("/user/calendar/item", check=401)
        self.mkcalendar("/user/calendar/item", check=401)

    def test_put_collections_and_items(self) -> None:
        """Test rights for creation of calendars and items with PUT."""
        self.configure({"auth": {"type": "none"}})
        self.put("/user/", "BEGIN:VCALENDAR\r\nEND:VCALENDAR", check=401)
        self.mkcol("/user/")
        self.put("/user/calendar/", "BEGIN:VCALENDAR\r\nEND:VCALENDAR")
        event1 = get_file_content("event1.ics")
        self.put("/user/calendar/event1.ics", event1)



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2020 Unrud <unrud@outlook.com>
# Copyright © 2024-2026 Peter Bieringer <pb@bieringer.de>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

import io
import logging
import re
import sys
import unicodedata
import xml.etree.ElementTree as ET
from typing import Optional, Union

from radicale import (auth, config, hook, httputils, log, pathutils, rights,
                      sharing, storage, types, utils, web, xmlutils)
from radicale.log import logger
from radicale.rights import intersect

# HACK: https://github.com/tiran/defusedxml/issues/54
import defusedxml.ElementTree as DefusedET  # isort:skip
sys.modules["xml.etree"].ElementTree = ET  # type:ignore[attr-defined]

USER_PATTERN_STRICT: str = "a-zA-Z0-9@\\.\\-_"
PATH_PATTERN_STRICT: str = USER_PATTERN_STRICT + "\\/~"  # / as separator

USER_PATTERN_STRICT_RE: str = "^[" + USER_PATTERN_STRICT + "]+$"
PATH_PATTERN_STRICT_RE: str = "^[" + PATH_PATTERN_STRICT + "]+$"

USER_BLACKLIST_MINIMAL: list = [":", "'", '"', '*', '?']
PATH_BLACKLIST_MINIMAL: list = USER_BLACKLIST_MINIMAL

BASE_WHITELIST_UNICODE: list = ["-", ".", "@", "_", "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "0"]

USER_WHITELIST_UNICODE: list = BASE_WHITELIST_UNICODE   # from USER_PATTERN_STRICT
PATH_WHITELIST_UNICODE: list = BASE_WHITELIST_UNICODE + ["/", "~"]  # from PATH_PATTERN_STRICT


def _check_format(self: storage.BaseStorage,
                  string: str,
                  blacklist_minimal: list[str],
                  whitelist_unicode: list[str],
                  validation_type: str,
                  ) -> bool:
    check_minimal = (validation_type == "minimal")
    check_unicode_letter = (validation_type == "unicode-letter")
    check_no_unicode = (validation_type == "no-unicode")
    logger.trace("_check_format investigate %r (validation_type=%r check_minimal=%s check_unicode_letter=%s check_no_unicode=%s)", string, validation_type, check_minimal, check_unicode_letter, check_no_unicode)
    if not self._supports_trailing_whitespace and string.endswith(' '):
        return False
    for c in string:
        if c <= chr(31) or (c >= chr(127) and c <= chr(159)):
            # ASCII: control char
            return False
        if unicodedata.category(c)[0] == "C":
            # https://unicodeplus.com/category
            # Unicode: control
            return False
        if check_minimal or not self._supports_problematic_chars:
            if c in blacklist_minimal:
                logger.trace("_check_format found %r", c)
                return False
        if check_unicode_letter:
            if c not in whitelist_unicode:
                if unicodedata.category(c)[0] != "L":
                    logger.trace("_check_format found %r (%r)", c, unicodedata.category(c)[0])
                    return False
        if check_no_unicode:
            if ord(c) > 255:
                return False
    return True


def _check_user_format(self: storage.BaseStorage,
                       user: str,
                       validation_type: str
                       ) -> bool:
    if validation_type == "strict":
        return (re.search(USER_PATTERN_STRICT_RE, user) is not None)
    else:
        return _check_format(self,
                             user,
                             USER_BLACKLIST_MINIMAL,
                             USER_WHITELIST_UNICODE,
                             validation_type,
                             )


def _check_path_format(self: storage.BaseStorage,
                       path: str,
                       validation_type: str
                       ) -> bool:
    if validation_type == "strict":
        return (re.search(PATH_PATTERN_STRICT_RE, path) is not None)
    else:
        return _check_format(self,
                             path,
                             PATH_BLACKLIST_MINIMAL,
                             PATH_WHITELIST_UNICODE,
                             validation_type,
                             )


class ApplicationBase:

    configuration: config.Configuration
    _auth: auth.BaseAuth
    _storage: storage.BaseStorage
    _rights: rights.BaseRights
    _web: web.BaseWeb
    _sharing: sharing.BaseSharing
    _encoding: str
    _max_resource_size: int
    _max_vevent_rrule_occurrence: int
    _permit_delete_collection: bool
    _permit_overwrite_collection: bool
    _strict_preconditions: bool
    _validate_user_value: str
    _validate_path_format: str
    _hook: hook.BaseHook

    def __init__(self, configuration: config.Configuration) -> None:
        self.configuration = configuration
        self._auth = auth.load(configuration)
        self._storage = storage.load(configuration)
        self._rights = rights.load(configuration)
        self._web = web.load(configuration)
        self._sharing = sharing.load(configuration)
        self._encoding = configuration.get("encoding", "request")
        self._log_bad_put_request_content = configuration.get("logging", "bad_put_request_content")
        self._response_content_on_debug = configuration.get("logging", "response_content_on_debug")
        self._request_content_on_debug = configuration.get("logging", "request_content_on_debug")
        self._response_content_on_notice_condition = configuration.get("logging", "response_content_on_notice_condition")
        self._request_content_on_notice_condition = configuration.get("logging", "request_content_on_notice_condition")
        self._limit_content = configuration.get("logging", "limit_content")
        self._validate_user_value = configuration.get("server", "validate_user_value")
        self._validate_path_value = configuration.get("server", "validate_path_value")
        self._hook = hook.load(configuration)

    def _read_xml_request_body(self,
                               environ: types.WSGIEnviron,
                               request_info: dict,
                               ) -> Optional[ET.Element]:
        content = httputils.decode_request(
            self.configuration, environ,
            httputils.read_raw_request_body(self.configuration, environ))
        if not content:
            return None
        try:
            xml_content = DefusedET.fromstring(content)
        except ET.ParseError as e:
            logger.debug("Request content (Invalid XML):\n%s", content)
            raise RuntimeError("Failed to parse XML: %s" % e) from e
        if self._request_content_on_debug:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Request content (XML):\n%s", utils.textwrap_str(xmlutils.pretty_xml(xml_content)))
        else:
            if self._request_content_on_notice_condition != {}:
                if log.log_conditional(
                        "request-content",
                        condition=self._request_content_on_notice_condition,
                        value=request_info,
                        ):
                    logger.notice("Request content (XML, log condition passed):\n%s", utils.textwrap_str(xmlutils.pretty_xml(xml_content)))
                else:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("Request content (XML, log condition skipped): suppressed")
            else:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("Request content (XML): suppressed by config/option [logging] request_content_on_debug")
        return xml_content

    def _xml_response(self, xml_content: ET.Element, request_info: dict) -> bytes:
        if self._response_content_on_debug:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Response content (XML):\n%s",
                             utils.textwrap_str(xmlutils.pretty_xml(xml_content), self._limit_content))
        else:
            if self._response_content_on_notice_condition != {}:
                if log.log_conditional(
                        "response-content",
                        condition=self._response_content_on_notice_condition,
                        value=request_info,
                        ):
                    logger.notice("Response content (XML, log condition passed):\n%s", utils.textwrap_str(xmlutils.pretty_xml(xml_content)))
                else:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("Response content (XML, log condition skipped): suppressed")
            else:
                logger.debug("Response content (XML): suppressed by config/option [logging] response_content_on_debug")
        f = io.BytesIO()
        ET.ElementTree(xml_content).write(f, encoding=self._encoding,
                                          xml_declaration=True)
        return f.getvalue()

    def _webdav_error_response(self, status: int, human_tag: str, request_info: dict
                               ) -> types.WSGIResponse:
        """Generate XML error response."""
        headers = {"Content-Type": "text/xml; charset=%s" % self._encoding}
        content = self._xml_response(xmlutils.webdav_error(human_tag), request_info)
        return status, headers, content, None


class Access:
    """Helper class to check access rights of an item"""

    user: str
    path: str
    parent_path: str
    permissions: str
    _rights: rights.BaseRights
    _parent_permissions: Optional[str]
    _permissions_filter: Union[str, None] = None

    def __init__(self, rights: rights.BaseRights, user: str, path: str, permissions_filter: Union[str, None] = None
                 ) -> None:
        self._rights = rights
        self.user = user
        self.path = path
        self.parent_path = pathutils.parent_path(path)
        self.permissions = self._rights.authorization(self.user, self.path)
        if permissions_filter is not None:
            self._permissions_filter = permissions_filter
            permissions_filtered = intersect(self.permissions, permissions_filter)
            self.permissions = permissions_filtered
        self._parent_permissions = None

    @property
    def parent_permissions(self) -> str:
        if self.path == self.parent_path:
            return self.permissions
        if self._parent_permissions is None:
            self._parent_permissions = self._rights.authorization(
                self.user, self.parent_path)
        if self._permissions_filter is not None:
            parent_permissions_filtered = intersect(self._parent_permissions, self._permissions_filter)
            self._parent_permissions = parent_permissions_filtered
        return self._parent_permissions

    def check(self, permission: str,
              item: Optional[types.CollectionOrItem] = None) -> bool:
        if permission not in "rwdDoO":
            raise ValueError("Invalid permission argument: %r" % permission)
        if not item:
            permissions = permission + permission.upper()
            parent_permissions = permission
        elif isinstance(item, storage.BaseCollection):
            if item.tag:
                permissions = permission
            else:
                permissions = permission.upper()
            parent_permissions = ""
        else:
            permissions = ""
            parent_permissions = permission
        return bool(rights.intersect(self.permissions, permissions) or (
            self.path != self.parent_path and
            rights.intersect(self.parent_permissions, parent_permissions)))



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2014 Jean-Marc Martins
# Copyright © 2012-2017 Guillaume Ayoub
# Copyright © 2017-2021 Unrud <unrud@outlook.com>
# Copyright © 2024-2025 Peter Bieringer <pb@bieringer.de>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

import os

from radicale import item as radicale_item
from radicale import pathutils, storage
from radicale.log import logger
from radicale.storage import multifilesystem
from radicale.storage.multifilesystem.base import StorageBase


class StoragePartMove(StorageBase):

    def move(self, item: radicale_item.Item,
             to_collection: storage.BaseCollection, to_href: str) -> None:
        if not pathutils.is_safe_filesystem_path_component(to_href):
            raise pathutils.UnsafePathError(to_href)
        assert isinstance(to_collection, multifilesystem.Collection)
        assert isinstance(item.collection, multifilesystem.Collection)
        assert item.href
        move_from = pathutils.path_to_filesystem(item.collection._filesystem_path, item.href, self._is_collision_free)
        move_to = pathutils.path_to_filesystem(to_collection._filesystem_path, to_href, self._is_collision_free)
        try:
            os.replace(move_from, move_to)
        except OSError as e:
            raise ValueError("Failed to move file %r => %r %s" % (move_from, move_to, e)) from e
        self._sync_directory(to_collection._filesystem_path)
        if item.collection._filesystem_path != to_collection._filesystem_path:
            self._sync_directory(item.collection._filesystem_path)
        # Move the item cache entry
        cache_folder = self._get_collection_cache_subfolder(item.collection._filesystem_path, ".Radicale.cache", "item")
        to_cache_folder = self._get_collection_cache_subfolder(to_collection._filesystem_path, ".Radicale.cache", "item")
        self._makedirs_synced(to_cache_folder)
        move_from = os.path.join(cache_folder, item.href)
        move_to = os.path.join(to_cache_folder, to_href)
        try:
            os.replace(move_from, move_to)
        except FileNotFoundError:
            pass
        except OSError as e:
            logger.error("Failed to move cache file %r => %r %s" % (move_from, move_to, e))
            pass
        else:
            self._makedirs_synced(to_cache_folder)
            if cache_folder != to_cache_folder:
                self._makedirs_synced(cache_folder)
        # Track the change
        to_collection._update_history_etag(to_href, item)
        item.collection._update_history_etag(item.href, None)
        to_collection._clean_history()
        if item.collection._filesystem_path != to_collection._filesystem_path:
            item.collection._clean_history()



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2012-2017 Guillaume Ayoub
# Copyright © 2017-2022 Unrud <unrud@outlook.com>
# Copyright © 2024-2026 Peter Bieringer <pb@bieringer.de>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Tests for storage backends.

"""

import json
import logging
import os
import re
import shutil
import tempfile
from typing import ClassVar, cast

import pytest

import radicale.tests.custom.storage_simple_sync
from radicale import logger, pathutils
from radicale.tests import BaseTest
from radicale.tests.helpers import get_file_content
from radicale.tests.test_base import TestBaseRequests as _TestBaseRequests


class TestMultiFileSystem(BaseTest):
    """Tests for multifilesystem."""

    def setup_method(self) -> None:
        _TestBaseRequests.setup_method(cast(_TestBaseRequests, self))
        self.configure({"storage": {"type": "multifilesystem"}})

    def test_folder_creation(self) -> None:
        """Verify that the folder is created."""
        folder = os.path.join(self.colpath, "subfolder")
        self.configure({"storage": {"filesystem_folder": folder}})
        assert os.path.isdir(folder)

    def test_folder_creation_with_umask(self) -> None:
        """Verify that the folder is created with umask."""
        folder = os.path.join(self.colpath, "subfolder")
        self.configure({"storage": {"filesystem_folder": folder, "folder_umask": "0077"}})
        assert os.path.isdir(folder)

    def test_fsync(self) -> None:
        """Create a directory and file with syncing enabled."""
        self.configure({"storage": {"_filesystem_fsync": "True"}})
        self.mkcalendar("/calendar.ics/")

    def test_hook(self) -> None:
        """Run hook."""
        self.configure({"storage": {"hook": "mkdir %s" % os.path.join(
            "collection-root", "created_by_hook")}})
        self.mkcalendar("/calendar.ics/")
        self.propfind("/created_by_hook/")

    def test_hook_read_access(self) -> None:
        """Verify that hook is not run for read accesses."""
        self.configure({"storage": {"hook": "mkdir %s" % os.path.join(
            "collection-root", "created_by_hook")}})
        self.propfind("/")
        self.propfind("/created_by_hook/", check=404)

    @pytest.mark.skipif(not shutil.which("flock"), reason="flock command not found")
    def test_hook_storage_locked(self) -> None:
        """Verify that the storage is locked when the hook runs."""
        self.configure({"storage": {"hook": (
            "flock -n .Radicale.lock || exit 0; exit 1")}})
        self.mkcalendar("/calendar.ics/")

    def test_hook_principal_collection_creation(self) -> None:
        """Verify that the hooks runs when a new user is created."""
        self.configure({"storage": {"hook": "mkdir %s" % os.path.join(
            "collection-root", "created_by_hook")}})
        self.configure({"auth": {"type": "none"}})
        self.propfind("/", login="user:")
        self.propfind("/created_by_hook/")

    def test_hook_fail(self) -> None:
        """Verify that a request succeeded if the hook still fails (anyhow no rollback implemented)."""
        self.configure({"storage": {"hook": "exit 1"}})
        self.mkcalendar("/calendar.ics/", check=201)

    def test_item_cache_rebuild(self) -> None:
        """Delete the item cache and verify that it is rebuild."""
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event1.ics")
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        _, answer1 = self.get(path)
        cache_folder = os.path.join(self.colpath, "collection-root",
                                    "calendar.ics", ".Radicale.cache", "item")
        assert os.path.exists(os.path.join(cache_folder, "event1.ics"))
        shutil.rmtree(cache_folder)
        _, answer2 = self.get(path)
        assert answer1 == answer2
        assert os.path.exists(os.path.join(cache_folder, "event1.ics"))

    def test_item_cache_rebuild_subfolder(self) -> None:
        """Delete the item cache and verify that it is rebuild."""
        self.configure({"storage": {"use_cache_subfolder_for_item": "True"}})
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event1.ics")
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        _, answer1 = self.get(path)
        cache_folder = os.path.join(self.colpath, "collection-cache",
                                    "calendar.ics", ".Radicale.cache", "item")
        assert os.path.exists(os.path.join(cache_folder, "event1.ics"))
        shutil.rmtree(cache_folder)
        _, answer2 = self.get(path)
        assert answer1 == answer2
        assert os.path.exists(os.path.join(cache_folder, "event1.ics"))

    def test_item_cache_rebuild_mtime_and_size(self) -> None:
        """Delete the item cache and verify that it is rebuild."""
        self.configure({"storage": {"use_mtime_and_size_for_item_cache": "True"}})
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event1.ics")
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        _, answer1 = self.get(path)
        cache_folder = os.path.join(self.colpath, "collection-root",
                                    "calendar.ics", ".Radicale.cache", "item")
        assert os.path.exists(os.path.join(cache_folder, "event1.ics"))
        shutil.rmtree(cache_folder)
        _, answer2 = self.get(path)
        assert answer1 == answer2
        assert os.path.exists(os.path.join(cache_folder, "event1.ics"))

    def test_put_items_multiple(self) -> None:
        """Upload 2 items to calendar, check that collection inode number stays."""
        self.configure({"logging": {"response_content_on_debug": "False",
                                    "response_header_on_debug": "False",
                                    "request_content_on_debug": "False",
                                    "request_header_on_debug": "False",
                                    }})
        self.mkcalendar("/calendar.ics/")
        event1 = get_file_content("event1.ics")
        path1 = "/calendar.ics/event1.ics"
        event2 = get_file_content("event2.ics")
        path2 = "/calendar.ics/event2.ics"
        self.put(path1, event1)
        collection_folder = os.path.join(self.colpath, "collection-root", "calendar.ics")
        collection_folder_st_ino_1 = os.stat(collection_folder).st_ino
        logger.debug("path=%r stat.ST_INO=%d", collection_folder, collection_folder_st_ino_1)
        self.put(path2, event2)
        collection_folder_st_ino_2 = os.stat(collection_folder).st_ino
        logger.debug("path=%r stat.ST_INO=%d", collection_folder, collection_folder_st_ino_2)
        assert collection_folder_st_ino_1 == collection_folder_st_ino_2

    def test_put_calendar_multiple(self) -> None:
        """Upload 2 calendars, check that collection inode number changes."""
        self.configure({"logging": {"response_content_on_debug": "False",
                                    "response_header_on_debug": "False",
                                    "request_content_on_debug": "False",
                                    "request_header_on_debug": "False",
                                    }})
        self.put("/calendar.ics/", "BEGIN:VCALENDAR\r\nEND:VCALENDAR")
        collection_folder = os.path.join(self.colpath, "collection-root", "calendar.ics")
        collection_folder_st_ino_1 = os.stat(collection_folder).st_ino
        logger.debug("path=%r stat.ST_INO=%d", collection_folder, collection_folder_st_ino_1)
        # Overwrite
        events = get_file_content("event_multiple.ics")
        self.put("/calendar.ics/", events)
        collection_folder_st_ino_2 = os.stat(collection_folder).st_ino
        logger.debug("path=%r stat.ST_INO=%d", collection_folder, collection_folder_st_ino_2)
        assert collection_folder_st_ino_1 != collection_folder_st_ino_2

    def test_put_whole_calendar_uids_used_as_file_names(self) -> None:
        """Test if UIDs are used as file names."""
        _TestBaseRequests.test_put_whole_calendar(
            cast(_TestBaseRequests, self))
        for uid in ("todo", "event"):
            _, answer = self.get("/calendar.ics/%s.ics" % uid)
            assert "\r\nUID:%s\r\n" % uid in answer

    def test_put_whole_calendar_random_uids_used_as_file_names(self) -> None:
        """Test if UIDs are used as file names."""
        _TestBaseRequests.test_put_whole_calendar_without_uids(
            cast(_TestBaseRequests, self))
        _, answer = self.get("/calendar.ics")
        assert answer is not None
        uids = []
        for line in answer.split("\r\n"):
            if line.startswith("UID:"):
                uids.append(line[len("UID:"):])
        for uid in uids:
            _, answer = self.get("/calendar.ics/%s.ics" % uid)
            assert answer is not None
            assert "\r\nUID:%s\r\n" % uid in answer

    def test_put_whole_addressbook_uids_used_as_file_names(self) -> None:
        """Test if UIDs are used as file names."""
        _TestBaseRequests.test_put_whole_addressbook(
            cast(_TestBaseRequests, self))
        for uid in ("contact1", "contact2"):
            _, answer = self.get("/contacts.vcf/%s.vcf" % uid)
            assert "\r\nUID:%s\r\n" % uid in answer

    def test_put_whole_addressbook_random_uids_used_as_file_names(
            self) -> None:
        """Test if UIDs are used as file names."""
        _TestBaseRequests.test_put_whole_addressbook_without_uids(
            cast(_TestBaseRequests, self))
        _, answer = self.get("/contacts.vcf")
        assert answer is not None
        uids = []
        for line in answer.split("\r\n"):
            if line.startswith("UID:"):
                uids.append(line[len("UID:"):])
        for uid in uids:
            _, answer = self.get("/contacts.vcf/%s.vcf" % uid)
            assert answer is not None
            assert "\r\nUID:%s\r\n" % uid in answer

    @pytest.mark.skipif(not pathutils.path_is_collision_free_case_sensitive(tempfile.mkdtemp()), reason="TEMP is not case sensitive")
    def test_collection_storage_dummy_case_sensitivity(self) -> None:
        """Test collection storage case sensitivity."""

    @pytest.mark.skipif(not pathutils.path_is_collision_free_no_short_filename(tempfile.mkdtemp()), reason="TEMP has short filename")
    def test_collection_storage_dummy_no_short_filename(self) -> None:
        """Test collection storage no short filename."""

    @pytest.mark.skipif(not pathutils.path_supports_unicode(tempfile.mkdtemp()), reason="TEMP is not supporting unicode")
    def test_collection_storage_dummy_no_support_of_unicode(self) -> None:
        """Test collection storage no support of unicode."""

    @pytest.mark.skipif(not pathutils.path_supports_trailing_whitespace(tempfile.mkdtemp()), reason="TEMP is not supporting trailing whitespace")
    def test_collection_storage_dummy_no_support_of_trailing_whitespace(self) -> None:
        """Test collection storage no support of trailing space."""

    @pytest.mark.skipif(not pathutils.path_supports_problematic_chars(tempfile.mkdtemp()), reason="TEMP is not supporting problematic chars")
    def test_collection_storage_dummy_no_support_of_problematic_chars(self) -> None:
        """Test collection storage no support of problematic chars."""

    @pytest.mark.skipif(not pathutils.path_supports_symlink(tempfile.mkdtemp()), reason="TEMP is not supporting symlink")
    def test_collection_sharing_by_softlink(self) -> None:
        """Test collection sharing by softlink."""
        self.configure({"auth": {"type": "none"}})
        path_group_col = "/group/calendar-shared.ics/"
        file_item = "event1.ics"
        self.mkcalendar(path_group_col, login="group:grouppw")
        event = get_file_content(file_item)
        self.put(path_group_col + file_item, event)
        fs_path_group_col = os.path.join(self.colpath, "collection-root", "group", "calendar-shared.ics")
        fs_path_group_col_rel = os.path.join("..", "group", "calendar-shared.ics")
        fs_path_user = os.path.join(self.colpath, "collection-root", "user")
        fs_path_user_col = os.path.join(self.colpath, "collection-root", "user", "calendar-group.ics")
        logger.debug("colpath=%r fs_path_group_col=%r", self.colpath, fs_path_group_col)
        assert os.path.isdir(fs_path_group_col)
        # create user directory and check
        logger.debug("create dir fs_path_user=%r", fs_path_user)
        os.mkdir(fs_path_user)
        assert os.path.isdir(fs_path_user)
        # create relative symlink and check
        logger.debug("create symlink src=%r dst=%r", fs_path_group_col_rel, fs_path_user_col)
        os.symlink(fs_path_group_col_rel, fs_path_user_col)
        logger.debug("test exists: fs_path_user_col=%r", fs_path_user_col)
        assert os.path.exists(fs_path_user_col)
        logger.debug("test is softlink: fs_path_user_col=%r", fs_path_user_col)
        assert os.path.islink(fs_path_user_col)
        self.propfind("/user/", login="user:userpw", HTTP_DEPTH="1")

    def test_colliding_items_by_file_case_insensitive(self, caplog) -> None:
        """Test for colliding files on file systems."""
        caplog.set_level(logging.WARNING)
        self.configure({"logging": {"request_content_on_debug": "False"}})
        fs_collision_free = pathutils.path_is_collision_free_case_sensitive(self.colpath)
        file_item = "EvEnT1.iCs"
        path_coll = "/calendar.ics/"
        self.mkcalendar(path_coll)
        event = get_file_content(file_item.lower())
        path_item = os.path.join(path_coll, file_item)
        path_uc = os.path.join(path_coll, file_item.upper())
        path_lc = os.path.join(path_coll, file_item.lower())
        self.put(path_item, event)
        self.put(path_uc, event, check=409)
        if not fs_collision_free:
            logs = caplog.messages
            assert len([log for log in logs if "File name collision" in log]) == 1
        self.put(path_lc, event, check=409)
        if not fs_collision_free:
            logs = caplog.messages
            assert len([log for log in logs if "File name collision" in log]) == 2

    def test_colliding_items_by_dir_case_insensitive(self, caplog) -> None:
        """Test for colliding dirs on file systems."""
        caplog.set_level(logging.WARNING)
        self.configure({"logging": {"request_content_on_debug": "False"}})
        fs_collision_free = pathutils.path_is_collision_free_case_sensitive(self.colpath)
        path_coll = "/CaLeNdAr.ics/"
        self.mkcalendar(path_coll)
        if fs_collision_free:
            self.mkcalendar(path_coll.lower(), check=201)
        else:
            self.mkcalendar(path_coll.lower(), check=409)
            logs = caplog.messages
            assert len([log for log in logs if "File name collision" in log]) == 1
        if fs_collision_free:
            self.mkcalendar(path_coll.upper(), check=201)
        else:
            self.mkcalendar(path_coll.upper(), check=409)
            logs = caplog.messages
            assert len([log for log in logs if "File name collision" in log]) == 2

    def test_colliding_items_by_dir_shortname(self, caplog) -> None:
        """Test for colliding dirs (shortname) on file systems."""
        caplog.set_level(logging.WARNING)
        self.configure({"logging": {"request_content_on_debug": "False"}})
        fs_collision_free = pathutils.path_is_collision_free_no_short_filename(self.colpath)
        fs_collision_free_case_sensitive = pathutils.path_is_collision_free_case_sensitive(self.colpath)
        path_coll = "/calendarlongname.ics/"
        path_coll_short = "/calend~1.ics/"
        self.mkcalendar(path_coll)
        if fs_collision_free:
            self.mkcalendar(path_coll_short.lower(), check=201)
            if not fs_collision_free_case_sensitive:
                # cleanup to avoid collision below
                self.delete(path_coll_short.lower())
        else:
            self.mkcalendar(path_coll_short.lower(), check=409)
            logs = caplog.messages
            assert len([log for log in logs if "File name collision" in log]) == 1
        if fs_collision_free:
            self.mkcalendar(path_coll_short.upper(), check=201)
        else:
            self.mkcalendar(path_coll_short.upper(), check=409)
            logs = caplog.messages
            assert len([log for log in logs if "File name collision" in log]) == 2

    @pytest.mark.skipif(not shutil.which("flock"), reason="flock command not found")
    @pytest.mark.skipif(radicale.log.logger.getEffectiveLevel() == logging.INFO, reason="requires loglevel DEBUG")
    def test_hook_placeholders_PUT(self, caplog) -> None:
        """Run hook and check placeholders via debug log: PUT"""
        caplog.set_level(logging.DEBUG)
        self.configure({"storage": {"hook": "echo \"hook-json {'user':'%(user)s', 'cwd':'%(cwd)s', 'path':'%(path)s', 'request':'%(request)s', 'to_path':'%(to_path)s'}\""}})
        found = 0
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event1.ics")
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        for line in caplog.messages:
            if line.find("\"hook-json ") != -1:
                found = 1
                r = re.search('.*\"hook-json ({.*})".*', line)
                if r:
                    s = r.group(1).replace("'", "\"")
                else:
                    break
                d = json.loads(s)
                if d["user"] == "Anonymous":
                    found = found | 2
                if d["cwd"]:
                    found = found | 4
                if d["path"]:
                    found = found | 8
                    if d["path"] == d["cwd"] + "/collection-root/calendar.ics/event1.ics":
                        found = found | 16
                if d["request"]:
                    found = found | 64
                    if d["request"] == "PUT":
                        found = found | 128
                if d["to_path"]:
                    found = found | 32
                    if d["to_path"] == "":
                        found = found | 256
                else:
                    found = found | 256 | 32
        if (found != 511):
            raise ValueError("Logging misses expected hook log line, found=%d data=%r", found, d)
        else:
            logging.info("Logging contains expected hook line, found=%d data=%r", found, d)

    @pytest.mark.skipif(not shutil.which("flock"), reason="flock command not found")
    @pytest.mark.skipif(radicale.log.logger.getEffectiveLevel() == logging.INFO, reason="requires loglevel DEBUG")
    def test_hook_placeholders_DELETE(self, caplog) -> None:
        """Run hook and check placeholders via debug log: DELETE"""
        caplog.set_level(logging.DEBUG)
        self.configure({"storage": {"hook": "echo \"hook-json {'user':'%(user)s', 'cwd':'%(cwd)s', 'path':'%(path)s', 'request':'%(request)s', 'to_path':'%(to_path)s'}\""}})
        found = 0
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event1.ics")
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        self.delete(path)
        for line in caplog.messages:
            if line.find("\"hook-json ") != -1:
                found = 1
                r = re.search('.*\"hook-json ({.*})".*', line)
                if r:
                    s = r.group(1).replace("'", "\"")
                else:
                    break
                d = json.loads(s)
                if d["user"] == "Anonymous":
                    found = found | 2
                if d["cwd"]:
                    found = found | 4
                if d["path"]:
                    found = found | 8
                    if d["path"] == d["cwd"] + "/collection-root/calendar.ics/event1.ics":
                        found = found | 16
                if d["request"]:
                    found = found | 64
                    if d["request"] == "DELETE":
                        found = found | 128
                if d["to_path"]:
                    found = found | 32
                    if d["to_path"] == "":
                        found = found | 256
                else:
                    found = found | 256 | 32
        if (found != 511):
            raise ValueError("Logging misses expected hook log line, found=%d data=%r", found, s)
        else:
            logging.info("Logging contains expected hook line, found=%d data=%r", found, d)

    @pytest.mark.skipif(not shutil.which("flock"), reason="flock command not found")
    @pytest.mark.skipif(radicale.log.logger.getEffectiveLevel() == logging.INFO, reason="requires loglevel DEBUG")
    def test_hook_placeholders_MKCALENDAR(self, caplog) -> None:
        """Run hook and check placeholders via debug log: MKCALENDAR"""
        caplog.set_level(logging.DEBUG)
        self.configure({"storage": {"hook": "echo \"hook-json {'user':'%(user)s', 'cwd':'%(cwd)s', 'path':'%(path)s', 'request':'%(request)s', 'to_path':'%(to_path)s'}\""}})
        found = 0
        self.mkcalendar("/calendar.ics/")
        for line in caplog.messages:
            if line.find("\"hook-json ") != -1:
                found = 1
                r = re.search('.*\"hook-json ({.*})".*', line)
                if r:
                    s = r.group(1).replace("'", "\"")
                else:
                    break
                d = json.loads(s)
                if d["user"] == "Anonymous":
                    found = found | 2
                if d["cwd"]:
                    found = found | 4
                if d["path"]:
                    found = found | 8
                    if d["path"] == d["cwd"] + "/collection-root/calendar.ics/":
                        found = found | 16
                if d["request"]:
                    found = found | 64
                    if d["request"] == "MKCALENDAR":
                        found = found | 128
                if d["to_path"]:
                    found = found | 32
                    if d["to_path"] == "":
                        found = found | 256
                else:
                    found = found | 256 | 32
        if (found != 511):
            raise ValueError("Logging misses expected hook log line, found=%d data=%r", found, d)
        else:
            logging.info("Logging contains expected hook line, found=%d data=%r", found, d)

    @pytest.mark.skipif(not shutil.which("flock"), reason="flock command not found")
    @pytest.mark.skipif(radicale.log.logger.getEffectiveLevel() == logging.INFO, reason="requires loglevel DEBUG")
    def test_hook_placeholders_MKCOL(self, caplog) -> None:
        """Run hook and check placeholders via debug log: MKCOL"""
        caplog.set_level(logging.DEBUG)
        self.configure({"storage": {"hook": "echo \"hook-json {'user':'%(user)s', 'cwd':'%(cwd)s', 'path':'%(path)s', 'request':'%(request)s', 'to_path':'%(to_path)s'}\""}})
        found = 0
        self.mkcol("/user1/")
        for line in caplog.messages:
            if line.find("\"hook-json ") != -1:
                found = 1
                r = re.search('.*\"hook-json ({.*})".*', line)
                if r:
                    s = r.group(1).replace("'", "\"")
                else:
                    break
                d = json.loads(s)
                if d["user"] == "Anonymous":
                    found = found | 2
                if d["cwd"]:
                    found = found | 4
                if d["path"]:
                    found = found | 8
                    if d["path"] == d["cwd"] + "/collection-root/user1/":
                        found = found | 16
                if d["request"]:
                    found = found | 64
                    if d["request"] == "MKCOL":
                        found = found | 128
                if d["to_path"]:
                    found = found | 32
                    if d["to_path"] == "":
                        found = found | 256
                else:
                    found = found | 256 | 32
        if (found != 511):
            raise ValueError("Logging misses expected hook log line, found=%d data=%r", found, d)
        else:
            logging.info("Logging contains expected hook line, found=%d data=%r", found, d)

    @pytest.mark.skipif(not shutil.which("flock"), reason="flock command not found")
    @pytest.mark.skipif(radicale.log.logger.getEffectiveLevel() == logging.INFO, reason="requires loglevel DEBUG")
    def test_hook_placeholders_PROPPATCH(self, caplog) -> None:
        """Run hook and check placeholders via debug log: PROPPATCH"""
        caplog.set_level(logging.DEBUG)
        self.configure({"storage": {"hook": "echo \"hook-json {'user':'%(user)s', 'cwd':'%(cwd)s', 'path':'%(path)s', 'request':'%(request)s', 'to_path':'%(to_path)s'}\""}})
        found = 0
        self.mkcalendar("/calendar.ics/")
        proppatch = get_file_content("proppatch_set_calendar_color.xml")
        _, responses = self.proppatch("/calendar.ics/", proppatch)
        for line in caplog.messages:
            if line.find("\"hook-json ") != -1:
                found = 1
                r = re.search('.*\"hook-json ({.*})".*', line)
                if r:
                    s = r.group(1).replace("'", "\"")
                else:
                    break
                d = json.loads(s)
                if d["user"] == "Anonymous":
                    found = found | 2
                if d["cwd"]:
                    found = found | 4
                if d["path"]:
                    found = found | 8
                    if d["path"] == d["cwd"] + "/collection-root/calendar.ics/":
                        found = found | 16
                if d["request"]:
                    found = found | 64
                    if d["request"] == "PROPPATCH":
                        found = found | 128
                if d["to_path"]:
                    found = found | 32
                    if d["to_path"] == "":
                        found = found | 256
                else:
                    found = found | 256 | 32
        if (found != 511):
            raise ValueError("Logging misses expected hook log line, found=%d data=%r", found, d)
        else:
            logging.info("Logging contains expected hook line, found=%d data=%r", found, d)

    @pytest.mark.skipif(not shutil.which("flock"), reason="flock command not found")
    @pytest.mark.skipif(radicale.log.logger.getEffectiveLevel() == logging.INFO, reason="requires loglevel DEBUG")
    def test_hook_placeholders_MOVE(self, caplog) -> None:
        """Run hook and check placeholders via debug log: MOVE"""
        caplog.set_level(logging.DEBUG)
        self.configure({"storage": {"hook": "echo \"hook-json {'user':'%(user)s', 'cwd':'%(cwd)s', 'path':'%(path)s', 'request':'%(request)s', 'to_path':'%(to_path)s'}\""}})
        found = 0
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event1.ics")
        path1 = "/calendar.ics/event1.ics"
        path2 = "/calendar.ics/event2.ics"
        self.put(path1, event)
        self.request("MOVE", path1, check=201,
                     HTTP_DESTINATION="http://127.0.0.1/"+path2)
        for line in caplog.messages:
            if line.find("\"hook-json ") != -1:
                found = 1
                r = re.search('.*\"hook-json ({.*})".*', line)
                if r:
                    s = r.group(1).replace("'", "\"")
                else:
                    break
                d = json.loads(s)
                if d["user"] == "Anonymous":
                    found = found | 2
                if d["cwd"]:
                    found = found | 4
                if d["path"]:
                    found = found | 8
                    if d["path"] == d["cwd"] + "/collection-root/calendar.ics/event1.ics":
                        found = found | 16
                if d["request"]:
                    found = found | 64
                    if d["request"] == "MOVE":
                        found = found | 128
                if d["to_path"]:
                    found = found | 32
                    if d["to_path"] == d["cwd"] + "/collection-root/calendar.ics/event2.ics":
                        found = found | 256
        if (found != 511):
            raise ValueError("Logging misses expected hook log line, found=%d data=%r", found, d)
        else:
            logging.info("Logging contains expected hook line, found=%d data=%r", found, d)


class TestMultiFileSystemNoLock(BaseTest):
    """Tests for multifilesystem_nolock."""

    def setup_method(self) -> None:
        _TestBaseRequests.setup_method(cast(_TestBaseRequests, self))
        self.configure({"storage": {"type": "multifilesystem_nolock"}})

    test_add_event = _TestBaseRequests.test_add_event
    test_item_cache_rebuild = TestMultiFileSystem.test_item_cache_rebuild


class TestCustomStorageSystem(BaseTest):
    """Test custom backend loading."""

    def setup_method(self) -> None:
        _TestBaseRequests.setup_method(cast(_TestBaseRequests, self))
        self.configure({"storage": {
            "type": "radicale.tests.custom.storage_simple_sync"}})

    full_sync_token_support: ClassVar[bool] = False

    test_add_event = _TestBaseRequests.test_add_event
    _report_sync_token = _TestBaseRequests._report_sync_token
    # include tests related to sync token
    s: str = ""
    for s in dir(_TestBaseRequests):
        if s.startswith("test_") and "sync" in s.split("_"):
            locals()[s] = getattr(_TestBaseRequests, s)
    del s


class TestCustomStorageSystemCallable(BaseTest):
    """Test custom backend loading with ``callable``."""

    def setup_method(self) -> None:
        _TestBaseRequests.setup_method(cast(_TestBaseRequests, self))
        self.configure({"storage": {
            "type": radicale.tests.custom.storage_simple_sync.Storage}})

    test_add_event = _TestBaseRequests.test_add_event



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2008 Nicolas Kandel
# Copyright © 2008 Pascal Halter
# Copyright © 2008-2017 Guillaume Ayoub
# Copyright © 2017-2019 Unrud <unrud@outlook.com>
# Copyright © 2024-2026 Peter Bieringer <pb@bieringer.de>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Authentication backend that checks credentials with a htpasswd file.

Apache's htpasswd command (httpd.apache.org/docs/programs/htpasswd.html)
manages a file for storing user credentials. It can encrypt passwords using
different the methods BCRYPT/SHA256/SHA512 or MD5-APR1 (a version of MD5 modified for
Apache). MD5-APR1 provides medium security as of 2015. Only BCRYPT/SHA256/SHA512 can be
considered secure by current standards.

MD5-APR1-encrypted credentials can be written by all versions of htpasswd (it
is the default, in fact), whereas BCRYPT/SHA256/SHA512 requires htpasswd 2.4.x or newer.

The `is_authenticated(user, password)` function provided by this module
verifies the user-given credentials by parsing the htpasswd credential file
pointed to by the ``htpasswd_filename`` configuration value while assuming
the password encryption method specified via the ``htpasswd_encryption``
configuration value.

The following htpasswd password encryption methods are supported by Radicale
out-of-the-box:
    - plain-text (created by htpasswd -p ...) -- INSECURE
    - MD5-APR1   (htpasswd -m ...) -- htpasswd's default method, INSECURE
    - SHA256     (htpasswd -2 ...)
    - SHA512     (htpasswd -5 ...)

When bcrypt is installed:
    - BCRYPT     (htpasswd -B ...) -- Requires htpasswd 2.4.x

When argon2 is installed:
    - ARGON2     (python -c 'from passlib.hash import argon2; print(argon2.using(type="ID").hash("password"))')

"""

import functools
import hmac
import os
import re
import threading
import time
from typing import Any, Tuple

from passlib.hash import apr_md5_crypt, sha256_crypt, sha512_crypt

from radicale import auth, config, logger


class Auth(auth.BaseAuth):

    BCRYPT_MAX_PWLEN = 72

    _filename: str
    _encoding: str
    _htpasswd: dict             # login -> digest
    _htpasswd_mtime_ns: int
    _htpasswd_size: int
    _htpasswd_ok: bool
    _htpasswd_not_ok_time: float
    _htpasswd_not_ok_reminder_seconds: int
    _htpasswd_bcrypt_use: int
    _htpasswd_argon2_use: int
    _htpasswd_cache: bool
    _has_bcrypt: bool
    _has_argon2: bool
    _encryption: str
    _lock: threading.Lock

    def __init__(self, configuration: config.Configuration) -> None:
        super().__init__(configuration)
        self._filename = configuration.get("auth", "htpasswd_filename")
        logger.info("auth htpasswd file: %r", self._filename)
        self._encoding = configuration.get("encoding", "stock")
        logger.info("auth htpasswd file encoding: %r", self._encoding)
        self._htpasswd_cache = configuration.get("auth", "htpasswd_cache")
        logger.info("auth htpasswd cache: %s", self._htpasswd_cache)
        self._encryption: str = configuration.get("auth", "htpasswd_encryption")
        logger.info("auth htpasswd encryption is 'radicale.auth.htpasswd_encryption.%s'", self._encryption)

        self._has_bcrypt = False
        self._has_argon2 = False
        self._htpasswd_ok = False
        self._htpasswd_not_ok_reminder_seconds = 60  # currently hardcoded
        (self._htpasswd_ok, self._htpasswd_bcrypt_use, self._htpasswd_argon2_use, self._htpasswd, self._htpasswd_size, self._htpasswd_mtime_ns) = self._read_htpasswd(True, False)
        self._lock = threading.Lock()

        if self._encryption == "plain":
            self._verify = self._plain
        elif self._encryption == "md5":
            self._verify = self._md5apr1
        elif self._encryption == "sha256":
            self._verify = self._sha256
        elif self._encryption == "sha512":
            self._verify = self._sha512

        if self._encryption == "bcrypt" or self._encryption == "autodetect":
            try:
                import bcrypt
            except ImportError as e:
                if (self._encryption == "autodetect") and (self._htpasswd_bcrypt_use == 0):
                    logger.warning("auth htpasswd encryption is 'radicale.auth.htpasswd_encryption.%s' which can require bycrypt module, but currently no entries found", self._encryption)
                else:
                    raise RuntimeError(
                        "The htpasswd encryption method 'bcrypt' or 'autodetect' requires "
                        "the bcrypt module (entries found: %d)." % self._htpasswd_bcrypt_use) from e
            else:
                self._has_bcrypt = True
                if self._encryption == "autodetect":
                    if self._htpasswd_bcrypt_use == 0:
                        logger.info("auth htpasswd encryption is 'radicale.auth.htpasswd_encryption.%s' and bcrypt module found, but currently not required", self._encryption)
                    else:
                        logger.info("auth htpasswd encryption is 'radicale.auth.htpasswd_encryption.%s' and bcrypt module found (bcrypt entries found: %d)", self._encryption, self._htpasswd_bcrypt_use)
            if self._encryption == "bcrypt":
                self._verify = functools.partial(self._bcrypt, bcrypt)
            else:
                self._verify = self._autodetect
                if self._htpasswd_bcrypt_use:
                    self._verify_bcrypt = functools.partial(self._bcrypt, bcrypt)

        if self._encryption == "argon2" or self._encryption == "autodetect":
            try:
                import argon2
                from passlib.hash import argon2  # noqa: F811
            except ImportError as e:
                if (self._encryption == "autodetect") and (self._htpasswd_argon2_use == 0):
                    logger.warning("auth htpasswd encryption is 'radicale.auth.htpasswd_encryption.%s' which can require argon2 module, but currently no entries found", self._encryption)
                else:
                    raise RuntimeError(
                        "The htpasswd encryption method 'argon2' or 'autodetect' requires "
                        "the argon2 module (entries found: %d)." % self._htpasswd_argon2_use) from e
            else:
                self._has_argon2 = True
                if self._encryption == "autodetect":
                    if self._htpasswd_argon2_use == 0:
                        logger.info("auth htpasswd encryption is 'radicale.auth.htpasswd_encryption.%s' and argon2 module found, but currently not required", self._encryption)
                    else:
                        logger.info("auth htpasswd encryption is 'radicale.auth.htpasswd_encryption.%s' and argon2 module found (argon2 entries found: %d)", self._encryption, self._htpasswd_argon2_use)
            if self._encryption == "argon2":
                self._verify = functools.partial(self._argon2, argon2)
            else:
                self._verify = self._autodetect
                if self._htpasswd_argon2_use:
                    self._verify_argon2 = functools.partial(self._argon2, argon2)

        if not hasattr(self, '_verify'):
            raise RuntimeError("The htpasswd encryption method %r is not "
                               "supported." % self._encryption)

    def _plain(self, hash_value: str, password: str) -> tuple[str, bool]:
        """Check if ``hash_value`` and ``password`` match, plain method."""
        return ("PLAIN", hmac.compare_digest(hash_value.encode(), password.encode()))

    def _plain_fallback(self, method_orig, hash_value: bytes, password: bytes) -> tuple[str, bool]:
        """Check if ``hash_value`` and ``password`` match, plain method / fallback in case of hash length is not matching on autodetection."""
        info = "PLAIN/fallback as hash length not matching for " + method_orig + ": " + str(len(hash_value))
        return (info, hmac.compare_digest(hash_value, password))

    def _bcrypt(self, bcrypt: Any, hash_value: str, password: str) -> tuple[str, bool]:
        pw = password.encode()
        hv = hash_value.encode()
        if len(pw) > self.BCRYPT_MAX_PWLEN:
            pw = pw[:self.BCRYPT_MAX_PWLEN]
            logger.warning("Bcrypt passwords can not be longer than %d characters, truncated" % self.BCRYPT_MAX_PWLEN)
        if self._encryption == "autodetect" and len(hv) != 60:
            return self._plain_fallback("BCRYPT", hv, pw)
        else:
            return ("BCRYPT", bcrypt.checkpw(password=pw, hashed_password=hv))

    def _argon2(self, argon2: Any, hash_value: str, password: str) -> tuple[str, bool]:
        return ("ARGON2", argon2.verify(password, hash_value.strip()))

    def _md5apr1(self, hash_value: str, password: str) -> tuple[str, bool]:
        return ("MD5-APR1", apr_md5_crypt.verify(password, hash_value.strip()))

    def _sha256(self, hash_value: str, password: str) -> tuple[str, bool]:
        return ("SHA-256", sha256_crypt.verify(password, hash_value.strip()))

    def _sha512(self, hash_value: str, password: str) -> tuple[str, bool]:
        return ("SHA-512", sha512_crypt.verify(password, hash_value.strip()))

    def _autodetect(self, hash_value: str, password: str) -> tuple[str, bool]:
        if re.match(r"^\$apr1\$[A-Za-z0-9/.]{8}\$[A-Za-z0-9/.]{22}", hash_value):
            # MD5-APR1
            return self._md5apr1(hash_value, password)
        elif re.match(r"^\$2(a|b|x|y)?\$[0-9]{2}\$[A-Za-z0-9/.]{53}", hash_value):
            # BCRYPT
            return self._verify_bcrypt(hash_value, password)
        elif re.match(r"^\$argon2(i|d|id)\$", hash_value):
            # ARGON2
            return self._verify_argon2(hash_value, password)
        elif re.match(r"^\$5\$(rounds=[0-9]+\$)?[A-Za-z0-9/.]{16}\$[A-Za-z0-9/.]{42}", hash_value):
            # SHA-256
            return self._sha256(hash_value, password)
        elif re.match(r"^\$6\$(rounds=[0-9]+\$)?[A-Za-z0-9/.]{16}\$[A-Za-z0-9/.]{85}", hash_value):
            # SHA-512
            return self._sha512(hash_value, password)
        else:
            return self._plain(hash_value, password)

    def _read_htpasswd(self, init: bool, suppress: bool) -> Tuple[bool, int, int, dict, int, int]:
        """Read htpasswd file

        init == True: stop on error
        init == False: warn/skip on error and set mark to log reminder every interval
        suppress == True: suppress warnings, change info to debug (used in non-caching mode)
        suppress == False: do not suppress warnings (used in caching mode)

        """
        htpasswd_ok = True
        bcrypt_use = 0
        argon2_use = 0
        if (init is True) or (suppress is True):
            info = "Read"
        else:
            info = "Re-read"
        if suppress is False:
            logger.info("%s content of htpasswd file start: %r", info, self._filename)
        else:
            logger.debug("%s content of htpasswd file start: %r", info, self._filename)
        htpasswd: dict[str, str] = dict()
        entries = 0
        duplicates = 0
        errors = 0
        try:
            with open(self._filename, encoding=self._encoding) as f:
                line_num = 0
                for line in f:
                    line_num += 1
                    line = line.rstrip("\n")
                    if line.lstrip() and not line.lstrip().startswith("#"):
                        try:
                            login, digest = line.split(":", maxsplit=1)
                            skip = False
                            if login == "" or digest == "":
                                if init is True:
                                    raise ValueError("htpasswd file contains problematic line not matching <login>:<digest> in line: %d" % line_num)
                                else:
                                    errors += 1
                                    logger.warning("htpasswd file contains problematic line not matching <login>:<digest> in line: %d (ignored)", line_num)
                                    htpasswd_ok = False
                                    skip = True
                            else:
                                if htpasswd.get(login):
                                    duplicates += 1
                                    if init is True:
                                        raise ValueError("htpasswd file contains duplicate login: '%s'", login, line_num)
                                    else:
                                        logger.warning("htpasswd file contains duplicate login: '%s' (line: %d / ignored)", login, line_num)
                                        htpasswd_ok = False
                                        skip = True
                                else:
                                    if re.match(r"^\$2(a|b|x|y)?\$", digest) and len(digest) == 60:
                                        if init is True:
                                            bcrypt_use += 1
                                        else:
                                            if self._has_bcrypt is False:
                                                logger.warning("htpasswd file contains bcrypt digest login: '%s' (line: %d / ignored because module is not loaded)", login, line_num)
                                                skip = True
                                                htpasswd_ok = False
                                    if re.match(r"^\$argon2(i|d|id)\$", digest):
                                        if init is True:
                                            argon2_use += 1
                                        else:
                                            if self._has_argon2 is False:
                                                logger.warning("htpasswd file contains argon2 digest login: '%s' (line: %d / ignored because module is not loaded)", login, line_num)
                                                skip = True
                                                htpasswd_ok = False
                            if skip is False:
                                htpasswd[login] = digest
                                entries += 1
                        except ValueError as e:
                            if init is True:
                                raise RuntimeError("Invalid htpasswd file %r: %s" % (self._filename, e)) from e
        except OSError as e:
            if init is True:
                raise RuntimeError("Failed to load htpasswd file %r: %s" % (self._filename, e)) from e
            else:
                logger.warning("Failed to load htpasswd file on re-read: %r" % self._filename)
                htpasswd_ok = False
        htpasswd_size = os.stat(self._filename).st_size
        htpasswd_mtime_ns = os.stat(self._filename).st_mtime_ns
        if suppress is False:
            logger.info("%s content of htpasswd file done: %r (entries: %d, duplicates: %d, errors: %d)", info, self._filename, entries, duplicates, errors)
        else:
            logger.debug("%s content of htpasswd file done: %r (entries: %d, duplicates: %d, errors: %d)", info, self._filename, entries, duplicates, errors)
        if htpasswd_ok is True:
            self._htpasswd_not_ok_time = 0
        else:
            self._htpasswd_not_ok_time = time.time()
        return (htpasswd_ok, bcrypt_use, argon2_use, htpasswd, htpasswd_size, htpasswd_mtime_ns)

    def _login(self, login: str, password: str) -> str:
        """Validate credentials.

        Iterate through htpasswd credential file until login matches, extract
        hash (encrypted password) and check hash against password,
        using the method specified in the Radicale config.

        Optional: the content of the file is cached and live updates will be detected by
        comparing mtime_ns and size

        """
        login_ok = False
        digest: str
        if self._htpasswd_cache is True:
            # check and re-read file if required
            with self._lock:
                htpasswd_size = os.stat(self._filename).st_size
                htpasswd_mtime_ns = os.stat(self._filename).st_mtime_ns
                if (htpasswd_size != self._htpasswd_size) or (htpasswd_mtime_ns != self._htpasswd_mtime_ns):
                    (self._htpasswd_ok, self._htpasswd_bcrypt_use, self._htpasswd_argon2_use, self._htpasswd, self._htpasswd_size, self._htpasswd_mtime_ns) = self._read_htpasswd(False, False)
                    self._htpasswd_not_ok_time = 0

            # log reminder of problemantic file every interval
            current_time = time.time()
            if (self._htpasswd_ok is False):
                if (self._htpasswd_not_ok_time > 0):
                    if (current_time - self._htpasswd_not_ok_time) > self._htpasswd_not_ok_reminder_seconds:
                        logger.warning("htpasswd file still contains issues (REMINDER, check warnings in the past): %r" % self._filename)
                        self._htpasswd_not_ok_time = current_time
                else:
                    self._htpasswd_not_ok_time = current_time

            if self._htpasswd.get(login):
                digest = self._htpasswd[login]
                login_ok = True
        else:
            # read file on every request
            (htpasswd_ok, htpasswd_bcrypt_use, htpasswd_argon2_use, htpasswd, htpasswd_size, htpasswd_mtime_ns) = self._read_htpasswd(False, True)
            if htpasswd.get(login):
                digest = htpasswd[login]
                login_ok = True

        if login_ok is True:
            try:
                (method, password_ok) = self._verify(digest, password)
            except ValueError as e:
                logger.error("Login verification failed for user: '%s' (htpasswd/%s) with error '%s'", login, self._encryption, e)
                return ""
            if password_ok:
                logger.debug("Login verification successful for user: '%s' (htpasswd/%s/%s)", login, self._encryption, method)
                return login
            else:
                logger.warning("Login verification failed for user: '%s' (htpasswd/%s/%s)", login, self._encryption, method)
        else:
            logger.warning("Login verification user not found (htpasswd): '%s'", login)
        return ""



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2026-2026 Max Berger <max@berger.name>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Integration tests for scene navigation (login/logout specific)
"""

import pathlib
from typing import Any, Generator

import pytest
from playwright.sync_api import Page, expect

from integ_tests.common import (NOSHARE_HTPASSWD, SHARING_HTPASSWD, Config,
                                login, start_radicale_server)


@pytest.fixture(params=[SHARING_HTPASSWD, NOSHARE_HTPASSWD], ids=lambda c: c.name)
def config(request: pytest.FixtureRequest) -> Config:
    return request.param


@pytest.fixture
def radicale_server(
    tmp_path: pathlib.Path, config: Config
) -> Generator[str, Any, None]:
    yield from start_radicale_server(tmp_path, config)


def test_login_logout_login(page: Page, radicale_server: str, config: Config) -> None:
    # 1. First login
    login(page, radicale_server, config)
    expect(page.locator("#collectionsscene")).to_be_visible()

    # 2. Logout
    page.click('#logoutview a[data-name="logout"]')
    expect(page.locator("#loginscene")).to_be_visible()
    expect(page.locator("#collectionsscene")).to_be_hidden()

    # 3. Second login
    login(page, radicale_server, config)
    expect(page.locator("#collectionsscene")).to_be_visible()


def test_login_wrong_password_shows_error(
    page: Page, radicale_server: str, config: Config
) -> None:
    page.goto(radicale_server)

    # 1. Login with wrong password
    page.fill('#loginscene input[data-name="user"]', "admin")
    page.fill('#loginscene input[data-name="password"]', "wrongpassword")
    page.click('button:has-text("Next")')

    # Expect error message to be shown
    expect(page.locator('#loginscene [data-name="error"]')).to_be_visible()

    # 2. Login with right password
    page.fill('#loginscene input[data-name="password"]', "admi$pass#word")
    page.click('button:has-text("Next")')

    # Expect successful login
    expect(page.locator("#collectionsscene")).to_be_visible()

    # 3. Logout
    page.click('#logoutview a[data-name="logout"]')
    expect(page.locator("#loginscene")).to_be_visible()
    expect(page.locator("#collectionsscene")).to_be_hidden()

    # Expect error message is NOT shown
    expect(page.locator('#loginscene [data-name="error"]')).not_to_be_visible()



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2011-2017 Guillaume Ayoub
# Copyright © 2017-2023 Unrud <unrud@outlook.com>
# Copyright © 2024-2026 Peter Bieringer <pb@bieringer.de>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Functions to set up Python's logging facility for Radicale's WSGI application.

Log messages are sent to the first available target of:

  - Error stream specified by the WSGI server in "wsgi.errors"
  - ``sys.stderr``

"""

import contextlib
import io
import ipaddress
import logging
import os
import re
import socket
import struct
import sys
import threading
import time
from functools import partial, partialmethod
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Dict, Iterator,
                    Mapping, Optional, Tuple, Union, cast)

from radicale import types

LOGGER_NAME: str = "radicale"
LOGGER_FORMATS: Mapping[str, str] = {
    "verbose": "[%(asctime)s] [%(ident)s] [%(levelname)s] %(message)s",
    "journal": "[%(ident)s] [%(levelname)s] %(message)s",
}
DATE_FORMAT: str = "%Y-%m-%d %H:%M:%S %z"

LOG_CONDITION_TOKEN: dict = {"method": "str",
                             "path": "str",
                             "useragent": "str",
                             "host": "ipaddress",
                             "login": "str",
                             "status": "int",
                             }
LOG_CONDITION_CONDITION: list = ["match", "value"]
LOG_CONDITION_MATCH_STR: list = ["startswith", "endswith", "equal", "re", "==", "="]
LOG_CONDITION_MATCH_INT: list = ["=", ">", "<", "<=", ">=", "equal", "==", "<>", "!="]
LOG_CONDITION_MATCH_IPADDRESS: list = ["==", "equal", "=", "<>", "!="]
LOG_CONDITION_MATCH_IPNETWORK: list = ["included", "excluded", "incl", "excl"]

LOG_LEVEL_OPTIONS: list = ["trace", "debug", "info", "notice", "warning", "error", "critical", "alert"]

LOG_LEVEL_TRACE: int = 5
LOG_LEVEL_NOTICE: int = 25
LOG_LEVEL_ALERT: int = 55

logging.addLevelName(LOG_LEVEL_TRACE, "TRACE")
logging.addLevelName(LOG_LEVEL_NOTICE, "NOTICE")
logging.addLevelName(LOG_LEVEL_ALERT, "ALERT")

setattr(logging, "TRACE", LOG_LEVEL_TRACE)
setattr(logging, "NOTICE", LOG_LEVEL_NOTICE)
setattr(logging, "ALERT", LOG_LEVEL_ALERT)

logging.__all__ += ['TRACE', 'NOTICE', 'ALERT']

setattr(logging, "trace", partial(logging.log, LOG_LEVEL_TRACE))
setattr(logging, "notice", partial(logging.log, LOG_LEVEL_NOTICE))
setattr(logging, "alert", partial(logging.log, LOG_LEVEL_ALERT))

setattr(logging.getLoggerClass(), "trace", partialmethod(logging.Logger.log, LOG_LEVEL_TRACE))
setattr(logging.getLoggerClass(), "notice", partialmethod(logging.Logger.log, LOG_LEVEL_NOTICE))
setattr(logging.getLoggerClass(), "alert", partialmethod(logging.Logger.log, LOG_LEVEL_ALERT))


class RadicaleLogger(logging.Logger):
    if TYPE_CHECKING:
        pass
        # reuse similar types
        trace = logging.Logger.debug
        notice = logging.Logger.info
        alert = logging.Logger.critical
    else:
        def trace(self, msg, *args, **kwargs):
            if self.isEnabledFor(LOG_LEVEL_TRACE):
                self._log(LOG_LEVEL_TRACE, msg, args, **kwargs)

        def notice(self, msg, *args, **kwargs):
            if self.isEnabledFor(LOG_LEVEL_NOTICE):
                self._log(LOG_LEVEL_NOTICE, msg, args, **kwargs)

        def alert(self, msg, *args, **kwargs):
            if self.isEnabledFor(LOG_LEVEL_ALERT):
                self._log(LOG_LEVEL_ALERT, msg, args, **kwargs)


logger = cast(RadicaleLogger, logging.getLogger(LOGGER_NAME))


class RemoveTracebackFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        record.exc_info = None
        return True


class PassTRACETOKENFilter(logging.Filter):
    def __init__(self, trace_filter: str):
        super().__init__()
        self.trace_filter = trace_filter

    def filter(self, record: logging.LogRecord) -> bool:
        if record.levelno == LOG_LEVEL_TRACE:
            if record.msg.startswith(self.trace_filter):
                return True
            else:
                return False
        else:
            return True


REMOVE_TRACEBACK_FILTER: logging.Filter = RemoveTracebackFilter()


class IdentLogRecordFactory:
    """LogRecordFactory that adds ``ident`` attribute."""

    def __init__(self, upstream_factory: Callable[..., logging.LogRecord]
                 ) -> None:
        self._upstream_factory = upstream_factory

    def __call__(self, *args: Any, **kwargs: Any) -> logging.LogRecord:
        record = self._upstream_factory(*args, **kwargs)
        ident = ("%d" % record.process if record.process is not None
                 else record.processName or "unknown")
        tid = None
        if record.thread is not None:
            if record.thread != threading.main_thread().ident:
                ident += "/%s" % (record.threadName or "unknown")
            if (sys.version_info >= (3, 8) and
                    record.thread == threading.get_ident()):
                try:
                    tid = threading.get_native_id()
                except AttributeError:
                    # so far function not existing e.g. on SunOS
                    # see also https://docs.python.org/3/library/threading.html#threading.get_native_id
                    tid = None

        record.ident = ident  # type:ignore[attr-defined]
        record.tid = tid  # type:ignore[attr-defined]
        return record


class ThreadedStreamHandler(logging.Handler):
    """Sends logging output to the stream registered for the current thread or
       ``sys.stderr`` when no stream was registered."""

    terminator: ClassVar[str] = "\n"

    _streams: Dict[int, types.ErrorStream]
    _journal_stream_id: Optional[Tuple[int, int]]
    _journal_socket: Optional[socket.socket]
    _journal_socket_failed: bool
    _formatters: Mapping[str, logging.Formatter]
    _formatter: Optional[logging.Formatter]

    def __init__(self, format_name: Optional[str] = None) -> None:
        super().__init__()
        self._streams = {}
        self._journal_stream_id = None
        with contextlib.suppress(TypeError, ValueError):
            dev, inode = os.environ.get("JOURNAL_STREAM", "").split(":", 1)
            self._journal_stream_id = (int(dev), int(inode))
        self._journal_socket = None
        self._journal_socket_failed = False
        self._formatters = {name: logging.Formatter(fmt, DATE_FORMAT)
                            for name, fmt in LOGGER_FORMATS.items()}
        self._formatter = (self._formatters[format_name]
                           if format_name is not None else None)

    def _get_formatter(self, default_format_name: str) -> logging.Formatter:
        return self._formatter or self._formatters[default_format_name]

    def _detect_journal(self, stream: types.ErrorStream) -> bool:
        if not self._journal_stream_id or not isinstance(stream, io.IOBase):
            return False
        try:
            stat = os.fstat(stream.fileno())
        except OSError:
            return False
        return self._journal_stream_id == (stat.st_dev, stat.st_ino)

    @staticmethod
    def _encode_journal(data: Mapping[str, Optional[Union[str, int]]]
                        ) -> bytes:
        msg = b""
        for key, value in data.items():
            if value is None:
                continue
            keyb = key.encode()
            valueb = str(value).encode()
            if b"\n" in valueb:
                msg += (keyb + b"\n" +
                        struct.pack("<Q", len(valueb)) + valueb + b"\n")
            else:
                msg += keyb + b"=" + valueb + b"\n"
        return msg

    def _try_emit_journal(self, record: logging.LogRecord) -> bool:
        if not self._journal_socket:
            # Try to connect to systemd journal socket
            if self._journal_socket_failed or not hasattr(socket, "AF_UNIX"):
                return False
            journal_socket = None
            try:
                journal_socket = socket.socket(
                    socket.AF_UNIX, socket.SOCK_DGRAM)
                journal_socket.connect("/run/systemd/journal/socket")
            except OSError as e:
                self._journal_socket_failed = True
                if journal_socket:
                    journal_socket.close()
                # Log after setting `_journal_socket_failed` to prevent loop!
                logger.error("Failed to connect to systemd journal: %s",
                             e, exc_info=True)
                return False
            self._journal_socket = journal_socket

        priority = {"TRACE": 7,
                    "DEBUG": 7,
                    "INFO": 6,
                    "NOTICE": 5,
                    "WARNING": 4,
                    "ERROR": 3,
                    "CRITICAL": 2,
                    "ALERT": 1}.get(record.levelname, 4)
        timestamp = time.strftime("%Y-%m-%dT%H:%M:%S.%%03dZ",
                                  time.gmtime(record.created)) % record.msecs
        data = {"PRIORITY": priority,
                "TID": cast(Optional[int], getattr(record, "tid", None)),
                "SYSLOG_IDENTIFIER": record.name,
                "SYSLOG_FACILITY": 1,
                "SYSLOG_PID": record.process,
                "SYSLOG_TIMESTAMP": timestamp,
                "CODE_FILE": record.pathname,
                "CODE_LINE": record.lineno,
                "CODE_FUNC": record.funcName,
                "MESSAGE": self._get_formatter("journal").format(record)}
        self._journal_socket.sendall(self._encode_journal(data))
        return True

    def emit(self, record: logging.LogRecord) -> None:
        try:
            stream = self._streams.get(threading.get_ident(), sys.stderr)
            if self._detect_journal(stream) and self._try_emit_journal(record):
                return
            msg = self._get_formatter("verbose").format(record)
            stream.write(msg + self.terminator)
            stream.flush()
        except Exception:
            self.handleError(record)

    @types.contextmanager
    def register_stream(self, stream: types.ErrorStream) -> Iterator[None]:
        """Register stream for logging output of the current thread."""
        key = threading.get_ident()
        self._streams[key] = stream
        try:
            yield
        finally:
            del self._streams[key]


@types.contextmanager
def register_stream(stream: types.ErrorStream) -> Iterator[None]:
    """Register stream for logging output of the current thread."""
    yield


def setup() -> None:
    """Set global logging up."""
    global register_stream
    format_name = os.environ.get("RADICALE_LOG_FORMAT") or None
    sane_format_name = format_name if format_name in LOGGER_FORMATS else None
    handler = ThreadedStreamHandler(sane_format_name)
    logging.basicConfig(handlers=[handler])
    register_stream = handler.register_stream
    log_record_factory = IdentLogRecordFactory(logging.getLogRecordFactory())
    logging.setLogRecordFactory(log_record_factory)
    set_level(logging.INFO, True)
    if format_name != sane_format_name:
        logger.error("Invalid RADICALE_LOG_FORMAT: %r", format_name)


logger_display_backtrace_disabled: bool = False
logger_display_backtrace_enabled: bool = False


def set_level(level: Union[int, str], backtrace_on_debug: bool, trace_filter: str = "") -> None:
    """Set logging level for global logger."""
    global logger_display_backtrace_disabled
    global logger_display_backtrace_enabled
    if isinstance(level, str):
        level = getattr(logging, level.upper())
        assert isinstance(level, int)
    logger.setLevel(level)
    logger.log(level, "Logging level set to: %r", logging.getLevelName(level))
    if level > logging.DEBUG:
        if logger_display_backtrace_disabled is False:
            logger.info("Logging of backtrace is disabled in this loglevel")
            logger_display_backtrace_disabled = True
        logger.addFilter(REMOVE_TRACEBACK_FILTER)
    else:
        if not backtrace_on_debug:
            if logger_display_backtrace_disabled is False:
                logger.debug("Logging of backtrace is disabled by option in this loglevel")
                logger_display_backtrace_disabled = True
            logger.addFilter(REMOVE_TRACEBACK_FILTER)
        else:
            if logger_display_backtrace_enabled is False:
                logger.debug("Logging of backtrace is enabled by option in this loglevel")
                logger_display_backtrace_enabled = True
            logger.removeFilter(REMOVE_TRACEBACK_FILTER)
    if level < logging.DEBUG:
        if trace_filter != "":
            logger.trace("Logging messages on 'trace' level enabled but filtered by prefix: %r", trace_filter)
            logger.addFilter(PassTRACETOKENFilter(trace_filter))
        else:
            logger.trace("Logging messages on 'trace' level enabled")


def log_conditional(name: str, condition: dict, value: dict) -> bool:
    logger.trace("log/conditional/%s/CHECK : condition=%r value=%r", name, condition, value)
    # "and" combination
    condition_active = False
    for token in condition:
        condition_active = True
        if LOG_CONDITION_TOKEN[token] == "str":
            logger.trace("log/conditional/%s/%s/string/CHECK  : %r %s %r", name, token, condition[token]["value"], condition[token]["match"], value[token])
            if condition[token]["match"] in ["equal", "==", "="]:
                if not condition[token]["value"] == value[token]:
                    return False
            elif condition[token]["match"] in ["startswith"]:
                if not value[token].startswith(condition[token]["value"]):
                    return False
            elif condition[token]["match"] in ["endswith"]:
                if not value[token].endswith(condition[token]["value"]):
                    return False
            elif condition[token]["match"] in ["re"]:
                if not re.search(condition[token]["value"], value[token]):
                    return False
            logger.trace("log/conditional/%s/%s/string/PASSED : %r %s %r", name, token, condition[token]["value"], condition[token]["match"], value[token])

        elif LOG_CONDITION_TOKEN[token] == "int":
            condition[token]["value"] = int(condition[token]["value"])
            logger.trace("log/conditional/%s/%s/integer/CHECK  : %r %s %r", name, token, condition[token]["value"], condition[token]["match"], value[token])
            if condition[token]["match"] in ["equal", "==", "="]:
                if not condition[token]["value"] == value[token]:
                    return False
            elif condition[token]["match"] in ["<"]:
                if not condition[token]["value"] < value[token]:
                    return False
            elif condition[token]["match"] in ["<="]:
                if not condition[token]["value"] <= value[token]:
                    return False
            elif condition[token]["match"] in [">"]:
                if not condition[token]["value"] > value[token]:
                    return False
            elif condition[token]["match"] in [">="]:
                if not condition[token]["value"] >= value[token]:
                    return False
            elif condition[token]["match"] in ["<>", "!="]:
                if not condition[token]["value"] != value[token]:
                    return False
            logger.trace("log/conditional/%s/%s/integer/PASSED : %r %s %r", name, token, condition[token]["value"], condition[token]["match"], value[token])

        elif LOG_CONDITION_TOKEN[token] == "ipaddress":
            logger.trace("log/conditional/%s/%s/ip/CHECK  : %r %s %r", name, token, condition[token]["value"], condition[token]["match"], value[token])
            # simple IP address check
            if condition[token]["match"] in ["equal", "==", "="]:
                if not condition[token]["value"] == value[token]:
                    return False
            elif condition[token]["match"] in ["<>", "!="]:
                if not condition[token]["value"] != value[token]:
                    return False
            elif condition[token]["match"] in ["included", "incl", "excluded", "excl"]:
                if condition[token]["match"] in ["included", "incl"]:
                    if value[token] == "unknown":
                        return False
                    else:
                        ip_net = ipaddress.ip_network(value[token])
                        ip = ipaddress.ip_network(condition[token]["value"])
                        if type(ip_net) is ipaddress.IPv4Network and type(ip) is ipaddress.IPv4Network:
                            if not ip_net.subnet_of(ip):
                                return False
                        elif type(ip_net) is ipaddress.IPv6Network and type(ip) is ipaddress.IPv6Network:
                            if not ip_net.subnet_of(ip):
                                return False
                        else:
                            return False
                elif condition[token]["match"] in ["excluded", "excl"]:
                    if value[token] != "unknown":
                        ip_net = ipaddress.ip_network(value[token])
                        if type(ip_net) is ipaddress.IPv4Network and type(ip) is ipaddress.IPv4Network:
                            if ip_net.subnet_of(ip):
                                return False
                        elif type(ip_net) is ipaddress.IPv6Network and type(ip) is ipaddress.IPv6Network:
                            if ip_net.subnet_of(ip):
                                return False
                logger.trace("log/conditional/%s/%s/ip/PASSED : %r %s %r", name, token, condition[token]["value"], condition[token]["match"], value[token])
            else:
                logger.trace("log/conditional/%s/%s/ip/SKIPPED: %r %s %r", name, token, condition[token]["value"], condition[token]["match"], value[token])

    # only return 'True' if at least one condition was found
    logger.trace("log/conditional/%s/PASSED: condition=%r value=%r", name, condition, value)
    return condition_active



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2026-2026 Max Berger <max@berger.name>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Integration tests for sharing (login/logout specific)
"""

import pathlib
import re
from typing import Any, Generator

import pytest
from playwright.sync_api import Page, expect

from integ_tests.common import (SHARING_HTPASSWD,
                                SHARING_HTPASSWD_USERSWITHDOMAIN, Config,
                                create_collection, login,
                                start_radicale_server)


@pytest.fixture(params=[SHARING_HTPASSWD, SHARING_HTPASSWD_USERSWITHDOMAIN])
def radicale_server_config(request: pytest.FixtureRequest) -> Config:
    return request.param


@pytest.fixture
def radicale_server(
    tmp_path: pathlib.Path, radicale_server_config: Config
) -> Generator[str, Any, None]:
    yield from start_radicale_server(tmp_path, radicale_server_config)


@pytest.mark.parametrize("permissions", ["ro", "rw"])
def test_incoming_shares(
    page: Page, radicale_server: str, radicale_server_config: Config, permissions: str
) -> None:
    # 1. Admin logs in and creates a map share for 'max'
    login(page, radicale_server, radicale_server_config)
    create_collection(page, radicale_server)

    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)
    page.click('button[data-name="sharebymap"]')
    page.locator('input[data-name="shareuser"]').fill(
        radicale_server_config.user_username
    )
    page.locator('input[data-name="sharehref"]').fill("mapped")
    if permissions == "rw":
        page.check("#newshare_attr_permissions_rw")
    page.click('#createeditsharescene button[data-name="submit"]')
    expect(
        page.locator("tr[data-name='sharemaprowtemplate']:not(.hidden)")
    ).to_have_count(1)
    page.click('#sharecollectionscene button[data-name="cancel"]')

    # 2. Admin logs out
    page.click('a[data-name="logout"]')

    # 3. Max logs in
    page.fill(
        '#loginscene input[data-name="user"]', radicale_server_config.user_username
    )
    page.fill('#loginscene input[data-name="password"]', "userpassword")
    page.click('button:has-text("Next")')

    # 4. Max sees the incoming share
    page.click('a[data-name="incomingshares"]')
    expect(page.locator("#incomingsharingscene")).to_be_visible()
    expect(
        page.locator("tr[data-name='incomingsharerowtemplate']:not(.hidden)")
    ).to_have_count(1)

    expect(
        page.locator(
            "tr[data-name='incomingsharerowtemplate']:not(.hidden) input[data-name='pathortoken']"
        )
    ).to_have_value(re.compile(r".*mapped/"))

    # 5. Max enables and shows the share
    # Initially, it's disabled and not shown (security by default)
    expect(
        page.locator(
            "tr[data-name='incomingsharerowtemplate']:not(.hidden) input[data-name='enabled']"
        )
    ).not_to_be_checked()
    expect(
        page.locator(
            "tr[data-name='incomingsharerowtemplate']:not(.hidden) input[data-name='shown']"
        )
    ).not_to_be_checked()
    expect(
        page.locator(
            "tr[data-name='incomingsharerowtemplate']:not(.hidden) input[data-name='shown']"
        )
    ).to_be_disabled()

    # Enable it
    page.check(
        "tr[data-name='incomingsharerowtemplate']:not(.hidden) input[data-name='enabled']"
    )
    expect(
        page.locator(
            "tr[data-name='incomingsharerowtemplate']:not(.hidden) input[data-name='shown']"
        )
    ).not_to_be_disabled()

    # Show it
    page.check(
        "tr[data-name='incomingsharerowtemplate']:not(.hidden) input[data-name='shown']"
    )
    expect(
        page.locator(
            "tr[data-name='incomingsharerowtemplate']:not(.hidden) input[data-name='shown']"
        )
    ).to_be_checked()
    # This is needed due to a potential race condition.
    expect(
        page.locator(
            "tr[data-name='incomingsharerowtemplate']:not(.hidden) input[data-name='shown']"
        )
    ).not_to_be_disabled()

    # 6. Verify "shared by admin" and button visibility in the collection article
    page.click('#incomingsharingscene button[data-name="close"]')
    expect(page.locator("#incomingsharingscene")).to_be_hidden()

    article = page.locator("article:not(.hidden)").first
    expect(article.locator('[data-name="shared-by"]')).to_be_visible()
    expect(article.locator('[data-name="shared-by-owner"]')).to_have_text(
        radicale_server_config.admin_username
    )

    # Action buttons are only visible on mouseover
    article.hover()

    # Share and delete buttons should be hidden for all incoming shares
    expect(article.locator('a[data-name="share"]')).to_be_hidden()
    expect(article.locator('[data-name="shareoption"]')).to_be_hidden()
    expect(article.locator('a[data-name="delete"]')).to_be_hidden()

    # Edit button is visible if either data write or property write is allowed.
    # In the test environment, permit_properties_overlay is true, so it's always visible.
    expect(article.locator('a[data-name="edit"]')).to_be_visible()

    # 7. Assert no error was shown
    expect(page.locator('#incomingsharingscene span[data-name="error"]')).to_be_hidden()


def test_no_incoming_shares_message(
    page: Page, radicale_server: str, radicale_server_config: Config
) -> None:
    # 1. Max logs in
    page.goto(radicale_server)
    page.fill(
        '#loginscene input[data-name="user"]', radicale_server_config.user_username
    )
    page.fill('#loginscene input[data-name="password"]', "userpassword")
    page.click('button:has-text("Next")')

    # 2. Max goes to incoming shares scene
    page.click('a[data-name="incomingshares"]')
    expect(page.locator("#incomingsharingscene")).to_be_visible()

    # 3. Verify that the table is hidden and the message is visible
    expect(page.locator("#incomingsharingscene table")).to_be_hidden()
    expect(
        page.locator('#incomingsharingscene [data-name="nosharesmessage"]')
    ).to_be_visible()
    expect(
        page.locator('#incomingsharingscene [data-name="nosharesmessage"]')
    ).to_have_text("No incoming shares")

    page.click('#incomingsharingscene button[data-name="close"]')
    expect(page.locator("#incomingsharingscene")).to_be_hidden()



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2026-2026 Max Berger <max@berger.name>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.


"""
Test for checking if the collections are sorted correctly.
"""

import pathlib
import re
from typing import Any, Generator

import pytest
from playwright.sync_api import Page, expect

from integ_tests.common import SHARING_HTPASSWD, login, start_radicale_server


@pytest.fixture
def radicale_server(tmp_path: pathlib.Path) -> Generator[str, Any, None]:
    yield from start_radicale_server(tmp_path, SHARING_HTPASSWD)


def create_named_collection(page: Page, name: str) -> None:
    page.click('.fabcontainer a[data-name="new"]')
    page.fill('#createcollectionscene input[data-name="displayname"]', name)
    page.click('#createcollectionscene button[data-name="submit"]')
    expect(page.locator("#createcollectionscene")).to_be_hidden()


def test_collection_sorting(page: Page, radicale_server: str) -> None:
    config = SHARING_HTPASSWD

    # 1. Admin logs in and creates "Z" and "A"
    login(page, radicale_server, config)
    create_named_collection(page, "Z")
    create_named_collection(page, "A")

    # 2. Admin shares "M" to "max"
    create_named_collection(page, "M")
    article_m = page.locator("article:not(.hidden)").filter(
        has=page.locator("[data-name='title']", has_text="M")
    )
    article_m.hover()
    article_m.locator("a[data-name='share']").click(force=True)

    page.click('button[data-name="sharebymap"]')
    page.locator('input[data-name="shareuser"]').fill(config.user_username)
    page.locator('input[data-name="sharehref"]').fill("m-shared")
    page.click('#createeditsharescene button[data-name="submit"]')
    page.click('#sharecollectionscene button[data-name="cancel"]')

    # 3. Admin logs out
    page.click('a[data-name="logout"]')

    # 4. Max logs in
    page.fill('#loginscene input[data-name="user"]', config.user_username)
    page.fill('#loginscene input[data-name="password"]', "userpassword")
    page.click('button:has-text("Next")')

    # 5. Max creates his own "B" and "Y"
    create_named_collection(page, "Y")
    create_named_collection(page, "B")

    # 6. Max enables the shared collection "M"
    page.click('a[data-name="incomingshares"]')
    row = page.locator("tr[data-name='incomingsharerowtemplate']:not(.hidden)")
    expect(row.locator("input[data-name='pathortoken']")).to_have_value(
        re.compile("m-shared")
    )
    row.locator("input[data-name='enabled']").check()
    row.locator("input[data-name='shown']").check()
    page.click('#incomingsharingscene button[data-name="close"]')

    # 7. Verify the order
    # Expected: "B", "Y" (Owned), then "M" (Shared)
    # Wait for articles to be rendered
    expect(page.locator("article:not(.hidden) [data-name='title']")).to_have_count(3)

    titles = page.locator(
        "article:not(.hidden) [data-name='title']"
    ).all_text_contents()
    titles = [t.strip() for t in titles if t.strip()]

    assert titles == ["B", "Y", "M"], f"Expected order ['B', 'Y', 'M'], got {titles}"

    # 8. Max creates "A"
    create_named_collection(page, "A")

    # Wait for articles to be rendered (now 4)
    expect(page.locator("article:not(.hidden) [data-name='title']")).to_have_count(4)

    titles = page.locator(
        "article:not(.hidden) [data-name='title']"
    ).all_text_contents()
    titles = [t.strip() for t in titles if t.strip()]

    # Expected: "A", "B", "Y", then "M"
    assert titles == [
        "A",
        "B",
        "Y",
        "M",
    ], f"Expected order ['A', 'B', 'Y', 'M'], got {titles}"



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2012-2017 Guillaume Ayoub
# Copyright © 2017-2018 Unrud <unrud@outlook.com>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Rights backend that allows authenticated users to read all calendars and
address books but only grants write access to their own.

"""

import radicale.rights.authenticated as authenticated
from radicale import pathutils


class Rights(authenticated.Rights):

    def authorization(self, user: str, path: str) -> str:
        if self._verify_user and not user:
            return ""
        sane_path = pathutils.strip_path(path)
        if not sane_path:
            return "R"
        if self._verify_user:
            owned = user == sane_path.split("/", maxsplit=1)[0]
        else:
            owned = True
        if "/" not in sane_path:
            return "RW" if owned else "R"
        if sane_path.count("/") == 1:
            return "rw" if owned else "r"
        return ""



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2025 Tobias Brox <tobias@tobix.eu>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Tests for pathutils module.

"""

import gc
import os
import tempfile

import pytest

from radicale import pathutils


class TestPathToFilesystem:
    """Tests for path_to_filesystem function."""

    @pytest.mark.filterwarnings("error::ResourceWarning")
    @pytest.mark.filterwarnings("error::pytest.PytestUnraisableExceptionWarning")
    def test_scandir_iterator_closed(self) -> None:
        """Verify that os.scandir iterator is properly closed.

        This test catches ResourceWarning: unclosed scandir iterator
        which occurs when os.scandir() is used without a context manager.
        See: https://github.com/Kozea/Radicale/issues/1972

        The ResourceWarning is emitted during garbage collection when an
        unclosed scandir iterator is finalized. We use pytest.mark.filterwarnings
        to convert both ResourceWarning and PytestUnraisableExceptionWarning
        to errors.
        """
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create a subdirectory so path_to_filesystem has something
            # to scan (the scandir check is for case-insensitive filesystems)
            subdir = os.path.join(tmpdir, "testdir")
            os.makedirs(subdir)

            # Call path_to_filesystem - if scandir iterator is not closed,
            # a ResourceWarning will be emitted during garbage collection
            result = pathutils.path_to_filesystem(tmpdir, "testdir")
            assert result == subdir

            # Force garbage collection to trigger any ResourceWarning
            # from unclosed iterators
            gc.collect()

    def test_path_to_filesystem_basic(self) -> None:
        """Test basic path_to_filesystem functionality."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Test empty path
            result = pathutils.path_to_filesystem(tmpdir, "")
            assert result == tmpdir

            # Test single component
            subdir = os.path.join(tmpdir, "test")
            os.makedirs(subdir)
            result = pathutils.path_to_filesystem(tmpdir, "test")
            assert result == subdir

            # Test nested path
            nested = os.path.join(subdir, "nested")
            os.makedirs(nested)
            result = pathutils.path_to_filesystem(tmpdir, "test/nested")
            assert result == nested

    def test_unsafe_path_raises(self) -> None:
        """Test that unsafe path components raise UnsafePathError."""
        with tempfile.TemporaryDirectory() as tmpdir:
            # Hidden files (starting with .) are not safe
            with pytest.raises(pathutils.UnsafePathError):
                pathutils.path_to_filesystem(tmpdir, ".hidden")

            # Backup files (ending with ~) are not safe
            with pytest.raises(pathutils.UnsafePathError):
                pathutils.path_to_filesystem(tmpdir, "backup~")



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2025-2025 Peter Bieringer <pb@bieringer.de>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Radicale tests related to hook 'email'

"""

import logging
import os
import re
from datetime import datetime, timedelta

from radicale.tests import BaseTest
from radicale.tests.helpers import get_file_content


class TestHooks(BaseTest):
    """Tests with hooks."""

    def setup_method(self) -> None:
        BaseTest.setup_method(self)
        rights_file_path = os.path.join(self.colpath, "rights")
        with open(rights_file_path, "w") as f:
            f.write("""\
[permit delete collection]
user: .*
collection: test-permit-delete
permissions: RrWwD

[forbid delete collection]
user: .*
collection: test-forbid-delete
permissions: RrWwd

[permit overwrite collection]
user: .*
collection: test-permit-overwrite
permissions: RrWwO

[forbid overwrite collection]
user: .*
collection: test-forbid-overwrite
permissions: RrWwo

[allow all]
user: .*
collection: .*
permissions: RrWw""")
        self.configure({"rights": {"file": rights_file_path,
                                   "type": "from_file"}})
        self.configure({"hook": {"type": "email",
                                 "dryrun": "True"}})

    def _future_date_timestamp(self) -> str:
        """Return a date timestamp for a future date."""
        future_date = datetime.now() + timedelta(days=1)
        return future_date.strftime("%Y%m%dT%H%M%S")

    def _future_date(self) -> str:
        """Return a date for a future date."""
        future_date = datetime.now() + timedelta(days=1)
        return future_date.strftime("%Y%m%d")

    def _past_date_timestamp(self) -> str:
        past_date = datetime.now() - timedelta(days=1)
        return past_date.strftime("%Y%m%dT%H%M%S")

    def _past_date(self) -> str:
        past_date = datetime.now() - timedelta(days=1)
        return past_date.strftime("%Y%m%d")

    def _replace_end_date_in_event(self, event: str, new_date: str) -> str:
        """Replace the end date in an event string."""
        return re.sub(r"DTEND;TZID=Europe/Paris:\d{8}T\d{6}",
                      f"DTEND;TZID=Europe/Paris:{new_date}", event)

    def _replace_end_onlydate_in_event(self, event: str, new_date: str) -> str:
        """Replace the end date in an event string."""
        return re.sub(r"DTEND;VALUE=DATE:\d{8}",
                      f"DTEND;VALUE=DATE:{new_date}", event)

    def test_add_event_with_future_end_date(self, caplog) -> None:
        caplog.set_level(logging.INFO)
        """Add an event."""
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event1.ics")
        event = self._replace_end_date_in_event(event, self._future_date_timestamp())
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        _, headers, answer = self.request("GET", path, check=200)
        assert "ETag" in headers
        assert headers["Content-Type"] == "text/calendar; charset=utf-8"
        assert "VEVENT" in answer
        assert "Event" in answer
        assert "UID:event" in answer

        logs = caplog.messages
        # Should have a log saying the notification item was received
        assert len([log for log in logs if "received notification_item: {'type': 'upsert'," in log]) == 1
        # Should NOT have a log saying that no email is sent (email won't actually be sent due to dryrun)
        assert len([log for log in logs if "New event detected, sending notifications to all attendees: event1" in log]) == 1

    def test_add_event_with_past_end_date(self, caplog) -> None:
        caplog.set_level(logging.WARNING)
        """Add an event."""
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event1.ics")
        event = self._replace_end_date_in_event(event, self._past_date_timestamp())
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        _, headers, answer = self.request("GET", path, check=200)
        assert "ETag" in headers
        assert headers["Content-Type"] == "text/calendar; charset=utf-8"
        assert "VEVENT" in answer
        assert "Event" in answer
        assert "UID:event" in answer

        logs = caplog.messages
        # Should have a log saying the notification item was received
        assert len([log for log in logs if "received notification_item: {'type': 'upsert'," in log]) == 1
        # Should have a log saying that no email is sent due to past end date
        assert len([log for log in logs if "Event end time is in the past, skipping notification for event: event1" in log]) == 1

    def test_delete_event_with_future_end_date(self, caplog) -> None:
        caplog.set_level(logging.INFO)
        """Delete an event."""
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event1.ics")
        event = self._replace_end_date_in_event(event, self._future_date_timestamp())
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        _, responses = self.delete(path)
        assert responses[path] == 200
        _, answer = self.get("/calendar.ics/")
        assert "VEVENT" not in answer

        logs = caplog.messages
        # Should have a log saying the notification item was received
        assert len([log for log in logs if "received notification_item: {'type': 'delete'," in log]) == 1
        # Should NOT have a log saying that no email is sent (email won't actually be sent due to dryrun)
        assert len([log for log in logs if "New event detected, sending notifications to all attendees: event1" in log]) == 1

    def test_delete_event_with_past_end_date(self, caplog) -> None:
        caplog.set_level(logging.WARNING)
        """Delete an event."""
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event1.ics")
        event = self._replace_end_date_in_event(event, self._past_date_timestamp())
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        _, responses = self.delete(path)
        assert responses[path] == 200
        _, answer = self.get("/calendar.ics/")
        assert "VEVENT" not in answer

        logs = caplog.messages
        # Should have a log saying the notification item was received
        assert len([log for log in logs if "received notification_item: {'type': 'delete'," in log]) == 1
        # Should have a log saying that no email is sent due to past end date
        assert len([log for log in logs if "Event end time is in the past, skipping notification for event: event1" in log]) == 1

    def test_delete_event_with_future_end_onlydate(self, caplog) -> None:
        caplog.set_level(logging.INFO)
        """Delete an event."""
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event_issue1917_1.ics")
        event = self._replace_end_onlydate_in_event(event, self._future_date())
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        _, responses = self.delete(path)
        assert responses[path] == 200
        _, answer = self.get("/calendar.ics/")
        assert "VEVENT" not in answer

        logs = caplog.messages
        # Should have a log saying the notification item was received
        assert len([log for log in logs if "received notification_item: {'type': 'delete'," in log]) == 1
        # Should NOT have a log saying that no email is sent (email won't actually be sent due to dryrun)
        assert len([log for log in logs if "New event detected, sending notifications to all attendees: event1" in log]) == 1

    def test_delete_event_with_past_onlyend_date(self, caplog) -> None:
        caplog.set_level(logging.WARNING)
        """Delete an event."""
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event_issue1917_1.ics")
        event = self._replace_end_onlydate_in_event(event, self._past_date())
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        _, responses = self.delete(path)
        assert responses[path] == 200
        _, answer = self.get("/calendar.ics/")
        assert "VEVENT" not in answer

        logs = caplog.messages
        # Should have a log saying the notification item was received
        assert len([log for log in logs if "received notification_item: {'type': 'delete'," in log]) == 1
        # Should have a log saying that no email is sent due to past end date
        assert len([log for log in logs if "Event end time is in the past, skipping notification for event: event1" in log]) == 1

    def test_add_event_with_future_mass1_end_date(self, caplog) -> None:
        self.configure({"hook": {"type": "email",
                                 "mass_email": "True",
                                 "dryrun": "True"}})

        caplog.set_level(logging.INFO)
        """Add an event."""
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event1.ics")
        event = self._replace_end_date_in_event(event, self._future_date_timestamp())
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        _, headers, answer = self.request("GET", path, check=200)
        assert "ETag" in headers
        assert headers["Content-Type"] == "text/calendar; charset=utf-8"
        assert "VEVENT" in answer
        assert "Event" in answer
        assert "UID:event" in answer

        logs = caplog.messages
        # Should have a log saying the notification item was received
        assert len([log for log in logs if "received notification_item: {'type': 'upsert'," in log]) == 1
        # Should NOT have a log saying that no email is sent (email won't actually be sent due to dryrun)
        assert len([log for log in logs if "New event detected, sending notifications to all attendees: event1" in log]) == 1
        assert len([log for log in logs if "Hello everyone" in log]) == 1

    def test_add_event_with_future_mass2_end_date(self, caplog) -> None:
        self.configure({"hook": {"type": "email",
                                 "mass_email": "True",
                                 "dryrun": "True"}})

        caplog.set_level(logging.INFO)
        """Add an event."""
        self.mkcalendar("/calendar.ics/")
        event = get_file_content("event1a1.ics")
        event = self._replace_end_date_in_event(event, self._future_date_timestamp())
        path = "/calendar.ics/event1.ics"
        self.put(path, event)
        _, headers, answer = self.request("GET", path, check=200)
        assert "ETag" in headers
        assert headers["Content-Type"] == "text/calendar; charset=utf-8"
        assert "VEVENT" in answer
        assert "Event" in answer
        assert "UID:event" in answer

        logs = caplog.messages
        # Should have a log saying the notification item was received
        assert len([log for log in logs if "received notification_item: {'type': 'upsert'," in log]) == 1
        # Should NOT have a log saying that no email is sent (email won't actually be sent due to dryrun)
        assert len([log for log in logs if "New event detected, sending notifications to all attendees: event1" in log]) == 1
        assert len([log for log in logs if "Hello everyone" in log]) == 0



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2026-2026 Max Berger <max@berger.name>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Integration tests for editing properties of a shared collection.
"""

import pathlib
import re
from typing import Any, Generator

import pytest
from playwright.sync_api import Page, expect

from integ_tests.common import SHARING_HTPASSWD, login, start_radicale_server


@pytest.fixture
def radicale_server(tmp_path: pathlib.Path) -> Generator[str, Any, None]:
    yield from start_radicale_server(tmp_path, SHARING_HTPASSWD)


def create_named_collection(page: Page, name: str) -> None:
    page.click('.fabcontainer a[data-name="new"]')
    page.fill('#createcollectionscene input[data-name="displayname"]', name)
    page.click('#createcollectionscene button[data-name="submit"]')
    expect(page.locator("#createcollectionscene")).to_be_hidden()


def test_shared_collection_property_edit(page: Page, radicale_server: str) -> None:
    config = SHARING_HTPASSWD

    # 1. Admin logs in and creates "Shared"
    login(page, radicale_server, config)
    create_named_collection(page, "Shared")

    # 2. Admin shares it with "max" with "Allow Properties write" enabled
    article = page.locator("article:not(.hidden)").filter(
        has=page.locator("[data-name='title']", has_text="Shared")
    )
    article.hover()
    article.locator("a[data-name='share']").click(force=True)

    page.click('button[data-name="sharebymap"]')
    page.locator('input[data-name="shareuser"]').fill(config.user_username)
    page.locator('input[data-name="sharehref"]').fill("shared-mapped")
    # Allow properties write
    page.check("#newshare_attr_properties_write_allow")
    page.click('#createeditsharescene button[data-name="submit"]')
    page.click('#sharecollectionscene button[data-name="cancel"]')

    # 3. Admin logs out
    page.click('a[data-name="logout"]')

    # 4. Max logs in
    page.fill('#loginscene input[data-name="user"]', config.user_username)
    page.fill('#loginscene input[data-name="password"]', "userpassword")
    page.click('button:has-text("Next")')

    # 5. Max enables the shared collection
    page.click('a[data-name="incomingshares"]')
    row = page.locator("tr[data-name='incomingsharerowtemplate']:not(.hidden)")
    expect(row.locator("input[data-name='pathortoken']")).to_have_value(
        re.compile("shared-mapped")
    )
    row.locator("input[data-name='enabled']").check()
    row.locator("input[data-name='shown']").check()
    page.click('#incomingsharingscene button[data-name="close"]')

    # 6. Verify "Edit" button is visible
    shared_article = page.locator("article:not(.hidden)").filter(
        has=page.locator("[data-name='title']", has_text="Shared")
    )
    shared_article.hover()
    expect(shared_article.locator("a[data-name='edit']")).to_be_visible()

    # 7. Max edits the collection
    shared_article.locator("a[data-name='edit']").click()
    page.fill('#editcollectionscene input[data-name="displayname"]', "Renamed by Max")
    page.click('#editcollectionscene button[data-name="submit"]')

    # 8. Verify the change
    expect(
        page.locator(
            "article:not(.hidden) [data-name='title']", has_text="Renamed by Max"
        )
    ).to_be_visible()



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2026-2026 Max Berger <max@berger.name>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Integration tests for sharing pages
"""

import pathlib
from typing import Any, Generator

import pytest
from playwright.sync_api import BrowserContext, Page, expect

from integ_tests.common import (SHARING_HTPASSWD, SHARING_XREMOTE, AuthType,
                                Config, SharingType, create_collection, login,
                                start_radicale_server)


@pytest.fixture(params=[SHARING_HTPASSWD, SHARING_XREMOTE], ids=lambda c: c.name)
def config(request: pytest.FixtureRequest) -> Config:
    return request.param


@pytest.fixture
def radicale_server(
    tmp_path: pathlib.Path, config: Config
) -> Generator[str, Any, None]:
    yield from start_radicale_server(tmp_path, config)


def test_create_and_delete_share_by_key(
    context: BrowserContext, page: Page, radicale_server: str, config: Config
) -> None:
    login(page, radicale_server, config, context=context)
    create_collection(page, radicale_server)
    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)

    expect(
        page.locator("tr[data-name='sharetokenrowtemplate']:not(.hidden)")
    ).to_have_count(0)

    page.click('button[data-name="sharebytoken"]')
    page.click('#createeditsharescene button[data-name="submit"]')
    expect(
        page.locator("tr[data-name='sharetokenrowtemplate']:not(.hidden)")
    ).to_have_count(1)
    expect(
        page.locator(
            "tr[data-name='sharetokenrowtemplate']:not(.hidden) span[data-name='ro']"
        )
    ).to_be_visible()
    expect(
        page.locator(
            "tr[data-name='sharetokenrowtemplate']:not(.hidden) span[data-name='rw']"
        )
    ).to_be_hidden()
    expect(
        page.locator(
            "tr[data-name='sharetokenrowtemplate']:not(.hidden) span[data-name='conversion']"
        )
    ).to_be_hidden()
    page.click('tr:not(.hidden) button[data-name="delete"]', strict=True)
    page.click('#deleteconfirmationscene button[data-name="delete"]')
    expect(
        page.locator("tr[data-name='sharetokenrowtemplate']:not(.hidden)")
    ).to_have_count(0)
    page.click('button[data-name="sharebytoken"]')
    page.click('label[for="newshare_attr_permissions_rw"]')
    page.click('#createeditsharescene button[data-name="submit"]')
    expect(
        page.locator("tr[data-name='sharetokenrowtemplate']:not(.hidden)")
    ).to_have_count(1)
    expect(
        page.locator(
            "tr[data-name='sharetokenrowtemplate']:not(.hidden) span[data-name='rw']"
        )
    ).to_be_visible()
    expect(
        page.locator(
            "tr[data-name='sharetokenrowtemplate']:not(.hidden) span[data-name='ro']"
        )
    ).to_be_hidden()
    expect(
        page.locator(
            "tr[data-name='sharetokenrowtemplate']:not(.hidden) span[data-name='conversion']"
        )
    ).to_be_hidden()
    page.click('tr:not(.hidden) button[data-name="delete"]', strict=True)
    page.click('#deleteconfirmationscene button[data-name="delete"]')
    expect(
        page.locator("tr[data-name='sharetokenrowtemplate']:not(.hidden)")
    ).to_have_count(0)


def test_create_and_delete_share_by_map(
    context: BrowserContext, page: Page, radicale_server: str, config: Config
) -> None:
    login(page, radicale_server, config, context=context)
    create_collection(page, radicale_server)
    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)

    expect(
        page.locator("tr[data-name='sharemaprowtemplate']:not(.hidden)")
    ).to_have_count(0)

    page.click('button[data-name="sharebymap"]')
    page.locator('input[data-name="shareuser"]').fill("max")
    page.locator('input[data-name="sharehref"]').fill("1234")
    page.click('#createeditsharescene button[data-name="submit"]')
    expect(
        page.locator("tr[data-name='sharemaprowtemplate']:not(.hidden)")
    ).to_have_count(1)
    expect(
        page.locator(
            "tr[data-name='sharemaprowtemplate']:not(.hidden) span[data-name='ro']"
        )
    ).to_be_visible()
    page.click('tr:not(.hidden) button[data-name="delete"]', strict=True)
    page.click('#deleteconfirmationscene button[data-name="delete"]')
    expect(
        page.locator("tr[data-name='sharemaprowtemplate']:not(.hidden)")
    ).to_have_count(0)
    page.click('button[data-name="sharebymap"]')
    page.click('label[for="newshare_attr_permissions_rw"]')
    page.locator('input[data-name="shareuser"]').fill("max")
    page.locator('input[data-name="sharehref"]').fill("1234")
    page.click('#createeditsharescene button[data-name="submit"]')
    expect(
        page.locator("tr[data-name='sharemaprowtemplate']:not(.hidden)")
    ).to_have_count(1)
    expect(
        page.locator(
            "tr[data-name='sharemaprowtemplate']:not(.hidden) span[data-name='rw']"
        )
    ).to_be_visible()
    page.click('tr:not(.hidden) button[data-name="delete"]', strict=True)
    page.click('#deleteconfirmationscene button[data-name="delete"]')
    expect(
        page.locator("tr[data-name='sharemaprowtemplate']:not(.hidden)")
    ).to_have_count(0)


def test_share_with_property_overrides(
    context: BrowserContext, page: Page, radicale_server: str, config: Config
) -> None:
    login(page, radicale_server, config, context=context)
    # Create a collection with specific details
    page.click('a[data-name="new"]')
    page.locator('#createcollectionscene input[data-name="displayname"]').fill(
        "Test Collection"
    )
    page.locator('#createcollectionscene input[data-name="description"]').fill(
        "Original Description"
    )
    page.locator('#createcollectionscene input[data-name="color"]').fill("#ff0000")
    page.click('#createcollectionscene button[data-name="submit"]')

    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)
    page.click('button[data-name="sharebytoken"]')

    # Verify property override is closed by default
    expect(
        page.locator('input[data-name="displayname_override_enabled"]')
    ).not_to_be_visible()
    page.click('details[data-name="properties_override"] summary')

    # Verify defaults
    expect(page.locator('input[data-name="displayname_override"]')).to_have_value(
        "Test Collection"
    )
    expect(page.locator('input[data-name="description_override"]')).to_have_value(
        "Original Description"
    )
    expect(page.locator('input[data-name="color_override"]')).to_have_value("#ff0000")
    expect(page.locator('input[data-name="displayname_override"]')).to_be_disabled()
    expect(page.locator('input[data-name="description_override"]')).to_be_disabled()
    expect(page.locator('input[data-name="color_override"]')).to_be_disabled()

    # Set overrides
    page.click('label[for="newshare_attr_displayname_enabled"]')
    page.locator('input[data-name="displayname_override"]').fill(
        "Overridden Displayname"
    )
    page.click('label[for="newshare_attr_description_enabled"]')
    page.locator('input[data-name="description_override"]').fill(
        "Overridden Description"
    )
    page.click('label[for="newshare_attr_color_enabled"]')
    page.locator('input[data-name="color_override"]').fill("#00ff00")

    page.click('#createeditsharescene button[data-name="submit"]')

    # Verify the share was created
    expect(
        page.locator("tr[data-name='sharetokenrowtemplate']:not(.hidden)")
    ).to_have_count(1)


def test_share_journal_no_overrides(
    context: BrowserContext, page: Page, radicale_server: str, config: Config
) -> None:
    login(page, radicale_server, config, context=context)
    # Create a collection of type JOURNAL
    page.click('a[data-name="new"]')
    page.locator('#createcollectionscene select[data-name="type"]').select_option(
        "JOURNAL"
    )
    page.locator('#createcollectionscene input[data-name="displayname"]').fill(
        "Test Journal"
    )
    page.locator('#createcollectionscene input[data-name="description"]').fill(
        "Journal Description"
    )
    page.click('#createcollectionscene button[data-name="submit"]')

    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)
    page.click('button[data-name="sharebytoken"]')

    # Verify property override visibility
    expect(page.locator('details[data-name="properties_override"]')).to_be_visible()
    expect(
        page.locator('input[data-name="displayname_override_enabled"]')
    ).not_to_be_visible()
    page.click('details[data-name="properties_override"] summary')

    expect(
        page.locator('input[data-name="displayname_override_enabled"]')
    ).to_be_visible()
    expect(
        page.locator('input[data-name="description_override_enabled"]')
    ).to_be_hidden()
    expect(page.locator('input[data-name="color_override_enabled"]')).to_be_hidden()

    # Create the share
    page.click('#createeditsharescene button[data-name="submit"]')

    # Verify the share was created
    expect(
        page.locator("tr[data-name='sharetokenrowtemplate']:not(.hidden)")
    ).to_have_count(1)


def test_edit_share_by_token(
    context: BrowserContext, page: Page, radicale_server: str, config: Config
) -> None:
    login(page, radicale_server, config, context=context)
    create_collection(page, radicale_server)
    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)

    # Create RO share
    page.click('button[data-name="sharebytoken"]')
    page.click('#createeditsharescene button[data-name="submit"]')
    expect(
        page.locator(
            "tr[data-name='sharetokenrowtemplate']:not(.hidden) span[data-name='ro']"
        )
    ).to_be_visible()

    # Edit to RW
    page.click('tr:not(.hidden) button[data-name="edit"]')
    expect(page.locator("#createeditsharescene h1")).to_have_text("Edit Share")
    page.click('label[for="newshare_attr_permissions_rw"]')
    page.click('#createeditsharescene button[data-name="submit"]')

    # Verify RW
    expect(
        page.locator(
            "tr[data-name='sharetokenrowtemplate']:not(.hidden) span[data-name='rw']"
        )
    ).to_be_visible()


def test_edit_share_by_map(
    context: BrowserContext, page: Page, radicale_server: str, config: Config
) -> None:
    login(page, radicale_server, config, context=context)
    create_collection(page, radicale_server)
    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)

    # Create RO map share
    page.click('button[data-name="sharebymap"]')
    page.locator('input[data-name="shareuser"]').fill("max")
    page.locator('input[data-name="sharehref"]').fill("mapped")
    page.click('#createeditsharescene button[data-name="submit"]')
    expect(
        page.locator(
            "tr[data-name='sharemaprowtemplate']:not(.hidden) span[data-name='ro']"
        )
    ).to_be_visible()

    # Edit map share
    page.click('tr:not(.hidden) button[data-name="edit"]')
    expect(page.locator("#createeditsharescene h1")).to_have_text("Edit Share")
    expect(page.locator('input[data-name="shareuser"]')).to_be_disabled()
    expect(page.locator('input[data-name="sharehref"]')).to_be_disabled()

    # Change permissions and enabled status
    page.click('label[for="newshare_attr_permissions_rw"]')
    page.uncheck('#createeditsharescene input[data-name="enabled"]')
    page.click('#createeditsharescene button[data-name="submit"]')

    # Verify changes
    expect(
        page.locator(
            "tr[data-name='sharemaprowtemplate']:not(.hidden) span[data-name='rw']"
        )
    ).to_be_visible()
    # If disabled, it might not show up or show differently, but our current UI doesn't visually distinguish enabled/disabled in the list yet
    # Let's verify by re-opening edit scene
    page.click('tr:not(.hidden) button[data-name="edit"]')
    expect(
        page.locator('#createeditsharescene input[data-name="enabled"]')
    ).not_to_be_checked()
    page.click('#createeditsharescene button[data-name="cancel"]')


def test_share_by_map_validation(
    context: BrowserContext, page: Page, radicale_server: str, config: Config
) -> None:
    login(page, radicale_server, config, context=context)
    create_collection(page, radicale_server)
    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)

    page.click('button[data-name="sharebymap"]')

    # Try empty user
    page.locator('input[data-name="shareuser"]').fill("")
    page.locator('input[data-name="sharehref"]').fill("1234")
    page.click('#createeditsharescene button[data-name="submit"]')
    expect(
        page.locator('#createeditsharescene [data-name="error"]:not(.hidden)')
    ).to_contain_text("Share User is empty")

    # Try logged in user
    page.locator('input[data-name="shareuser"]').fill("admin")
    page.click('#createeditsharescene button[data-name="submit"]')
    expect(
        page.locator('#createeditsharescene [data-name="error"]:not(.hidden)')
    ).to_contain_text("Share User cannot be admin")

    # Valid user
    page.locator('input[data-name="shareuser"]').fill("max")
    page.click('#createeditsharescene button[data-name="submit"]')

    # Verify success
    expect(
        page.locator("tr[data-name='sharemaprowtemplate']:not(.hidden)")
    ).to_have_count(1)


def test_create_and_delete_share_by_bday(
    context: BrowserContext, page: Page, radicale_server: str, config: Config
) -> None:
    login(page, radicale_server, config, context=context)
    # create collection of type ADDRESSBOOK for bday (bday only works with ADDRESSBOOK)
    page.click('a[data-name="new"]')
    page.locator('#createcollectionscene select[data-name="type"]').select_option(
        "ADDRESSBOOK"
    )
    page.locator('#createcollectionscene input[data-name="displayname"]').fill(
        "Addressbook For Bday"
    )
    page.click('#createcollectionscene button[data-name="submit"]')

    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)

    expect(
        page.locator("tr[data-name='sharemaprowtemplate']:not(.hidden)")
    ).to_have_count(0)

    page.click('button[data-name="sharebymap"]')
    page.click('label[for="newshare_conv_bday"]')

    # share to self (admin), which is allowed for conversions
    page.locator('input[data-name="shareuser"]').fill("admin")

    # verify that the permissions section is disabled
    expect(page.locator("input#newshare_attr_permissions_ro")).to_be_disabled()
    expect(page.locator("input#newshare_attr_permissions_rw")).to_be_disabled()

    page.click('#createeditsharescene button[data-name="submit"]')
    expect(
        page.locator("tr[data-name='sharemaprowtemplate']:not(.hidden)")
    ).to_have_count(1)

    # verify no permissions pill in the bday row
    expect(
        page.locator(
            "tr[data-name='sharemaprowtemplate']:not(.hidden) span[data-name='ro']"
        )
    ).to_be_hidden()
    expect(
        page.locator(
            "tr[data-name='sharemaprowtemplate']:not(.hidden) span[data-name='rw']"
        )
    ).to_be_hidden()

    # Instead, there should be the conversion icon
    expect(
        page.locator(
            "tr[data-name='sharemaprowtemplate']:not(.hidden) span[data-name='conversion']"
        )
    ).to_have_count(1)

    # Edit the bday share to verify that the conversion selection is disabled in the loaded state
    page.click(
        "tr[data-name='sharemaprowtemplate']:not(.hidden) button[data-name='edit']",
        strict=True,
    )
    expect(page.locator("#createeditsharescene h1")).to_have_text("Edit Share")
    expect(page.locator("input#newshare_conv_none")).to_be_disabled()
    expect(page.locator("input#newshare_conv_bday")).to_be_disabled()
    page.click('#createeditsharescene button[data-name="cancel"]')

    # Close the share scene and verify the virtual bday calendar is now in the collections list
    page.click('#sharecollectionscene button[data-name="cancel"]')
    expect(page.locator("#sharecollectionscene")).to_be_hidden()

    # The virtual calendar (bdaymapped) should appear as its own article
    # after the cache was invalidated following the self-share
    expect(page.locator("article:not(.hidden)")).to_have_count(2)

    # Delete the bday share by re-opening the share scene
    page.hover("article:not(.hidden) >> nth=0")
    page.click('article:not(.hidden) >> nth=0 >> a[data-name="share"]', force=True)
    page.click(
        "tr[data-name='sharemaprowtemplate']:not(.hidden) button[data-name='delete']",
        strict=True,
    )
    page.click('#deleteconfirmationscene button[data-name="delete"]')
    expect(
        page.locator("tr[data-name='sharemaprowtemplate']:not(.hidden)")
    ).to_have_count(0)


def test_bday_section_hidden_for_calendar(
    context: BrowserContext, page: Page, radicale_server: str, config: Config
) -> None:
    """Verify the bday calendar section is hidden for CALENDAR collections."""
    login(page, radicale_server, config, context=context)

    page.click('a[data-name="new"]')
    page.locator('#createcollectionscene select[data-name="type"]').select_option(
        "CALENDAR"
    )
    page.locator('#createcollectionscene input[data-name="displayname"]').fill(
        "My Calendar"
    )
    page.click('#createcollectionscene button[data-name="submit"]')

    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)

    expect(page.locator("#sharecollectionscene")).to_be_visible()
    page.click('button[data-name="sharebymap"]')
    expect(page.locator("details[data-name='conversions']")).to_be_hidden()
    page.click('#createeditsharescene button[data-name="cancel"]')


def test_bday_section_visible_for_addressbook(
    context: BrowserContext, page: Page, radicale_server: str, config: Config
) -> None:
    """Verify the bday calendar section is visible for ADDRESSBOOK collections."""
    login(page, radicale_server, config, context=context)

    page.click('a[data-name="new"]')
    page.locator('#createcollectionscene select[data-name="type"]').select_option(
        "ADDRESSBOOK"
    )
    page.locator('#createcollectionscene input[data-name="displayname"]').fill(
        "My Addressbook"
    )
    page.click('#createcollectionscene button[data-name="submit"]')

    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)

    expect(page.locator("#sharecollectionscene")).to_be_visible()
    page.click('button[data-name="sharebymap"]')
    expect(page.locator("details[data-name='conversions']")).to_be_visible()
    page.click('#createeditsharescene button[data-name="cancel"]')


@pytest.fixture
def map_disabled_server(tmp_path: pathlib.Path) -> Generator[str, Any, None]:
    config = Config(
        name="map_disabled",
        auth_type=AuthType.HTPASSWD,
        sharing_type=SharingType.NOSHARING,
        extra_config=f"""[sharing]
type = csv
collection_by_map = false
collection_by_token = true
permit_create_token = true
permit_create_map = false
database_path = {tmp_path / 'sharing.csv'}
""",
    )
    yield from start_radicale_server(tmp_path, config)


@pytest.fixture
def token_disabled_server(tmp_path: pathlib.Path) -> Generator[str, Any, None]:
    config = Config(
        name="token_disabled",
        auth_type=AuthType.HTPASSWD,
        sharing_type=SharingType.NOSHARING,
        extra_config=f"""[sharing]
type = csv
collection_by_map = true
collection_by_token = false
permit_create_token = false
permit_create_map = true
database_path = {tmp_path / 'sharing.csv'}
""",
    )
    yield from start_radicale_server(tmp_path, config)


@pytest.fixture
def both_disabled_server(tmp_path: pathlib.Path) -> Generator[str, Any, None]:
    config = Config(
        name="both_disabled",
        auth_type=AuthType.HTPASSWD,
        sharing_type=SharingType.NOSHARING,
        extra_config=f"""[sharing]
type = csv
collection_by_map = false
collection_by_token = false
permit_create_token = false
permit_create_map = false
database_path = {tmp_path / 'sharing.csv'}
""",
    )
    yield from start_radicale_server(tmp_path, config)


def test_map_sharing_disabled(
    context: BrowserContext, page: Page, map_disabled_server: str
) -> None:
    login(page, map_disabled_server, SHARING_HTPASSWD, context=context)
    create_collection(page, map_disabled_server)
    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)

    expect(
        page.locator('#sharecollectionscene button[data-name="sharebymap"]')
    ).to_be_hidden()
    expect(
        page.locator('#sharecollectionscene button[data-name="sharebytoken"]')
    ).to_be_visible()


def test_token_sharing_disabled(
    context: BrowserContext, page: Page, token_disabled_server: str
) -> None:
    login(page, token_disabled_server, SHARING_HTPASSWD, context=context)
    create_collection(page, token_disabled_server)
    page.hover("article:not(.hidden)")
    page.click('article:not(.hidden) a[data-name="share"]', force=True, strict=True)

    expect(
        page.locator('#sharecollectionscene button[data-name="sharebytoken"]')
    ).to_be_hidden()
    expect(
        page.locator('#sharecollectionscene button[data-name="sharebymap"]')
    ).to_be_visible()


def test_both_sharing_disabled(
    context: BrowserContext, page: Page, both_disabled_server: str
) -> None:
    login(page, both_disabled_server, SHARING_HTPASSWD, context=context)
    create_collection(page, both_disabled_server)
    page.hover("article:not(.hidden)")

    expect(page.locator('article:not(.hidden) a[data-name="share"]')).to_be_hidden()



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2008 Nicolas Kandel
# Copyright © 2008 Pascal Halter
# Copyright © 2008-2017 Guillaume Ayoub
# Copyright © 2017-2021 Unrud <unrud@outlook.com>
# Copyright © 2020-2020 Tom Hacohen <tom@stosb.com>
# Copyright © 2025-2026 Peter Bieringer <pb@bieringer.de>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

from radicale import httputils, types
from radicale.app.base import ApplicationBase


class ApplicationPartPost(ApplicationBase):

    def do_POST(self, environ: types.WSGIEnviron, base_prefix: str,
                path: str, user: str, request_info: dict) -> types.WSGIResponse:
        """Manage POST request."""
        if path == "/.web" or path.startswith("/.web/"):
            return self._web.post(environ, base_prefix, path, user, request_info)
        elif path == "/.sharing" or path.startswith("/.sharing/"):
            return self._sharing.post(environ, base_prefix, path, user, request_info)
        return httputils.METHOD_NOT_ALLOWED



# This file is part of Radicale - CalDAV and CardDAV server
# Copyright © 2025-2025 Peter Bieringer <pb@bieringer.de>
#
# This library is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radicale.  If not, see <http://www.gnu.org/licenses/>.

"""
Authentication backend that takes the username from the
``HTTP_REMOTE_USER`` header.

It's intended for use with a reverse proxy. Be aware as this will be insecure
if the reverse proxy is not configured properly.

"""

from typing import Tuple, Union

from radicale import types
from radicale.auth import none


class Auth(none.Auth):

    def get_external_login(self, environ: types.WSGIEnviron) -> Union[
            Tuple[()], Tuple[str, str]]:
        return environ.get("HTTP_REMOTE_USER", ""), ""
